import ast
from collections import namedtuple
from rich import print
from simpleeval import EvalWithCompoundTypes
from typing import Any, Union
from edsl.exceptions import (
    SurveyRuleCannotEvaluateError,
    SurveyRuleCollectionHasNoRulesAtNodeError,
    SurveyRuleRefersToFutureStateError,
    SurveyRuleReferenceInRuleToUnknownQuestionError,
    SurveyRuleSendsYouBackwardsError,
    SurveyRuleSkipLogicSyntaxError,
)
from edsl.surveys.base import EndOfSurvey
from edsl.utilities.ast_utilities import extract_variable_names
from edsl.utilities.interface import print_table_with_rich


class Rule:
    """
    The class defines a a "rule" for what question an agent should be presented next.

    The key structure of a rule is:
    - current_q: The question at which the rule is potentially applied (an index)
    - expression: A string that if evaluates to true, then next_q (an index) is next
    - next_q: The question is true
    - priority: an integer that determines which rule is applied if multiple rules apply

    The key part of this is expression specifiying the skip logic.
    An expression can be any combination of and/or/not/parenthesis etc.

            "q1 == 'yes' or q2 == 'no'"

    The skip-logic expression has to be about questions "before" the current node.
    To determine the value of the expression, we currently, we use eval.
    Eventually, we'll use the AST of the expression to make it safer.

    If multiple rules apply, the one with the highest priority is used.
    This is to deal with the fact that when we create a survey, we give the
    next question as the default.
    So when a question is added with index, it is always given a rule the next question is index + 1, but
    given a low (-1) priority.

    Only one rule should apply at each priority level.
    If there are conflicting rules, an exception is raised.
    Ideally, we'd have a way to resolve this ex ante, perhaps to traversing the implied tree
    each time a rule is added, but for now, we'll let the error emerge at run-time.

    ## Not implemented but nice to have
    We could potentially use the question pydantic models to check for rule conflicts, as
    they define the potential trees through a survey.

    We could also use the AST to check for conflicts by inspecting the types of a rule.
    For example, if we know the answer to a question is a string, we could check that
    the expression only contains string comparisons.
    This would be a lot of work.
    """

    def __init__(
        self,
        current_q: int,
        expression: str,
        next_q: Union[int, EndOfSurvey],
        question_name_to_index: dict[str, int],
        priority: int,
    ):
        """Questions are represented by int indices."""

        self.current_q = current_q
        self.expression = expression
        self.next_q = next_q
        self.priority = priority
        self.question_name_to_index = question_name_to_index

        if not isinstance(next_q, EndOfSurvey) and current_q > next_q:
            raise SurveyRuleSendsYouBackwardsError

        # get the AST for the expression - used to extract
        # the variables referenced in the expression
        try:
            self.ast_tree = ast.parse(self.expression)
        except SyntaxError:
            raise SurveyRuleSkipLogicSyntaxError

        # get the names of the variables in the expression
        # e.g., q1 == 'yes' -> ['q1']
        extracted_question_names = extract_variable_names(self.ast_tree)

        # make sure all the variables in the expression are known questions
        try:
            assert all([q in question_name_to_index for q in extracted_question_names])
        except AssertionError:
            raise SurveyRuleReferenceInRuleToUnknownQuestionError

        # get the indices of the questions mentioned in the expression
        self.named_questions_by_index = [
            question_name_to_index[q] for q in extracted_question_names
        ]

        # A rule should only refer to questions that have already been asked.
        # so the named questions in the expression should not be higher than the current question
        if self.named_questions_by_index:
            if max(self.named_questions_by_index) > self.current_q:
                raise SurveyRuleRefersToFutureStateError

    def to_dict(self):
        return {
            "current_q": self.current_q,
            "expression": self.expression,
            "next_q": "EndOfSurvey"
            if isinstance(self.next_q, EndOfSurvey)
            else self.next_q,
            "priority": self.priority,
            "question_name_to_index": self.question_name_to_index,
        }

    @classmethod
    def from_dict(self, rule_dict):
        if rule_dict["next_q"] == "EndOfSurvey":
            rule_dict["next_q"] = EndOfSurvey()

        return Rule(
            current_q=rule_dict["current_q"],
            expression=rule_dict["expression"],
            next_q=rule_dict["next_q"],
            priority=rule_dict["priority"],
            question_name_to_index=rule_dict["question_name_to_index"],
        )

    def __repr__(self):
        return f'Rule(current_q = {self.current_q}, expression = "{self.expression}", next_q = {self.next_q}, priority = {self.priority})'

    def __str__(self):
        return self.__repr__()

    @property
    def question_index_to_name(self):
        """Reverses the dictionary do we can look up questions by name"""
        return {v: k for k, v in self.question_name_to_index.items()}

    def show_ast_tree(self):
        """Pretty-prints the AST tree to the terminal"""
        print(
            ast.dump(
                self.ast_tree, annotate_fields=True, indent=4, include_attributes=True
            )
        )

    def evaluate(self, answers: dict[int, Any]):
        """
        Computes the value of the expression, given a dictionary of known questions answers.
        If the expression cannot be evaluated, it raises a CannotEvaluate exception.
        """

        def substitute_in_answers(expression, answers):
            "Take the dictionary of answers and substitute them into the expression"
            for var, value in answers.items():
                # If it's a string, add quotes; otherwise, just convert to string
                if isinstance(value, str):
                    replacement = f"'{value}'"
                else:
                    replacement = str(value)

                expression = expression.replace(var, replacement)
            return expression

        try:
            return EvalWithCompoundTypes().eval(
                substitute_in_answers(self.expression, answers)
            )
        except Exception as e:
            print(f"Exception in evaluation: {e}")
            raise SurveyRuleCannotEvaluateError


NextQuestion = namedtuple(
    "NextQuestion", "next_q, num_rules_found, expressions_evaluating_to_true, priority"
)


class RuleCollection:
    "A collection of rules for a particular survey"

    def __init__(self, rules=None, verbose=False):
        if rules is None:
            self.rules = []
        else:
            self.rules = rules
        self.verbose = verbose

    def __len__(self):
        return len(self.rules)

    def __getitem__(self, index):
        return self.rules[index]

    def __repr__(self):
        return f"RuleCollection(rules = {self.rules})"

    def to_dict(self):
        return [rule.to_dict() for rule in self.rules]

    @classmethod
    def from_dict(self, rule_collection_dict):
        return RuleCollection(
            rules=[Rule.from_dict(rule_dict) for rule_dict in rule_collection_dict]
        )

    def add_rule(self, rule: Rule):
        """Adds a rule to a survey. If it's not, return human-readable complaints"""
        self.rules.append(rule)

    def show_rules(self) -> None:
        if self.verbose:
            keys = ["current_q", "expression", "next_q", "priority"]
            rule_list = []
            for rule in sorted(self.rules, key=lambda r: r.current_q):
                rule_list.append({k: getattr(rule, k) for k in keys})

            print_table_with_rich(rule_list)

    def which_rules(self, q_now) -> list:
        """Which rules apply at the current node?
        More than one rule can apply. E.g., suppose we are at node 1.
        We could have three rules:
        1. "q1 == 'a' ==> 3
        2. "q1 == 'b' ==> 4
        3. "q1 == 'c' ==> 5
        """
        applicable_rules = [rule for rule in self.rules if rule.current_q == q_now]
        return applicable_rules

    def next_question(self, q_now, answers) -> int:
        "Find the next question by index, given the rule collection"
        # what rules apply at the current node?
        applicable_rules = self.which_rules(q_now)
        # Every node should have a rule - if it doesn't, there's a problem
        if not applicable_rules:
            raise SurveyRuleCollectionHasNoRulesAtNodeError

        # tracking
        expressions_evaluating_to_true = 0
        next_q = None
        highest_priority = -2  # start with -2 to 'pick up' the default rule added
        num_rules_found = 0

        for rule in applicable_rules:
            num_rules_found += 1
            try:
                if rule.evaluate(answers):  # evaluates to True
                    expressions_evaluating_to_true += 1
                    if rule.priority > highest_priority:  # higher priority
                        # we have a new champ!
                        next_q, highest_priority = rule.next_q, rule.priority
            except SurveyRuleCannotEvaluateError:
                pass

        return NextQuestion(
            next_q, num_rules_found, expressions_evaluating_to_true, highest_priority
        )

    def all_nodes_reachable(self) -> bool:
        """Are all nodes reachable, given rule set?
        To do the tree traversal, you'd have to instantiate answers to check.
        What would you do for continuous measures? Probably fork at above/below threshold?

        You could use the rules to help you draw the tree, no?
        Like you start linear
        Then take a rule, given it's conditions - figure out what options you need to track
        to build the tree e.g., if it mentions q1.answer, you need to add leafs at that node
        """
        raise NotImplementedError

    def no_cycles(self, rules: list[Rule]) -> bool:
        """Ensures there are not cycles in the rule set, which would create an infinite loop
        Probably also put a check in the Exam class to make sure they
        don't see the same question more than once - though
        that could be OK if their prompt text has changed.
        Python 3.10 has a topological sort, so
        """
        raise NotImplemented


if __name__ == "__main__":
    r = Rule(
        current_q=1,
        expression="q1 == 'yes'",
        next_q=2,
        question_name_to_index={"q1": 1},
        priority=0,
    )



class AgentErrors(Exception):
    pass


class AgentAttributeLookupCallbackError(AgentErrors):
    pass


class AgentCombinationError(AgentErrors):
    pass


class AgentLacksLLMError(AgentErrors):
    pass


class AgentRespondedWithBadJSONError(AgentErrors):
    pass



import time
from tqdm import tqdm
from edsl.jobs.JobsRunner import JobsRunner
from edsl.results import Result, Results


class JobsRunnerSerial(JobsRunner):
    """This JobRunner conducts interviews serially."""

    def run(
        self,
        n: int = 1,
        verbose: bool = False,
        sleep: int = 0,
        debug: bool = False,
        progress_bar: bool = True,
    ) -> Results:
        """
        Conducts Interviews **serially** and returns their results.
        - `n`: how many times to run each interview
        - `debug`: prints debug messages
        - `verbose`: prints messages
        - `progress_bar`: shows a progress bar
        """
        data = []
        for i in range(n):
            desc = "Running surveys" if n == 1 else f"Running surveys ({i+1}/{n})"
            interviews = (
                tqdm(self.interviews, desc=desc) if progress_bar else self.interviews
            )
            for interview in interviews:
                answer = interview.conduct_interview(debug=debug)
                result = Result(
                    agent=interview.agent,
                    scenario=interview.scenario,
                    model=interview.model,
                    iteration=i,
                    answer=answer,
                )
                data.append(result)
                time.sleep(sleep)

        return Results(survey=self.jobs.survey, data=data)



from __future__ import annotations
import json
import time
from abc import ABC, abstractmethod
from queue import Queue
from threading import Lock
from typing import Any, Callable, Type
from edsl.data import CRUDOperations, CRUD
from edsl.exceptions import LanguageModelResponseNotJSONError
from edsl.language_models.schemas import model_prices
from edsl.trackers.TrackerAPI import TrackerAPI


# from edsl.language_models.repair import repair


def repair(x):
    return x


class LanguageModel(ABC):
    """ABC for LLM subclasses."""

    def __init__(self, crud: CRUDOperations = CRUD, **kwargs):
        """
        Attributes:
        - all attributes inherited from subclasses
        - lock: lock for this model to ensure TODO
        - api_queue: queue that records messages about API calls the model makes. Used by `InterviewManager` to update details about state of model.
        """
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.lock = Lock()
        self.api_queue = Queue()
        self.crud = crud

    #######################
    # CORE METHODS
    #######################
    @abstractmethod
    def execute_model_call(
        self, prompt: str, system_prompt: str = ""
    ) -> dict[str, Any]:
        """Calls the LLM's API and returns the API response.
        This is an abstract method that needs to be implemented by the child class, on a
        model (and API-provider basis).

        For example, the GPT-4 model (as or right now)
        requires an API call that looks like this:

            openai.chat.completions.create(
                model=self.model,
                messages=[
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": prompt},
                ],
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                top_p=self.top_p,
                frequency_penalty=self.frequency_penalty,
                presence_penalty=self.presence_penalty,
            ).model_dump()

        It will then return JSON that needs to be parsed.
        This is also model/API specific.
        """
        raise NotImplementedError

    @abstractmethod
    def parse_response(raw_response: dict[str, Any]) -> str:
        """Parses the API response and returns the response text.
        What is returned by the API is model-specific and often includes meta-data that we do not need.
        For example, here is the results from a call to GPT-4:

        {
            "id": "chatcmpl-8eORaeuVb4po9WQRjKEFY6w7v6cTm",
            "choices": [
                {
                    "finish_reason": "stop",
                    "index": 0,
                    "logprobs": None,
                    "message": {
                        "content": "Hello! How can I assist you today? If you have any questions or need information on a particular topic, feel free to ask.",
                        "role": "assistant",
                        "function_call": None,
                        "tool_calls": None,
                    },
                }
            ],
            "created": 1704637774,
            "model": "gpt-4-1106-preview",
            "object": "chat.completion",
            "system_fingerprint": "fp_168383a679",
            "usage": {"completion_tokens": 27, "prompt_tokens": 13, "total_tokens": 40},
        }

        To actually tract the response, we need to grab
            data["choices[0]"]["message"]["content"].
        """
        raise NotImplementedError

    def _update_response_with_tracking(
        self, response, start_time, cached_response=False
    ):
        end_time = time.time()
        response["elapsed_time"] = end_time - start_time
        response["timestamp"] = end_time
        self._post_tracker_event(response)
        with self.lock:
            response["cached_response"] = cached_response
        return response

    def _get_raw_response(self, prompt: str, system_prompt: str = "") -> dict[str, Any]:
        """This is some middle-ware that handles the caching of responses.
        If the cache isn't being used, it just returns a 'fresh' call to the LLM,
        but appends some tracking information to the response (using the _update_response_with_tracking method).
        But if cache is being used, it first checks the database to see if the response is already there.
        If it is, it returns the cached response, but again appends some tracking information.
        If it isn't, it calls the LLM, saves the response to the database, and returns the response with tracking information.

        If self.use_cache is True, then attempts to retrieve the response from the database;
        if not in the DB, calls the LLM and writes the response to the DB."""
        start_time = time.time()

        if not self.use_cache:
            response = self.execute_model_call(prompt, system_prompt)
            return self._update_response_with_tracking(response, start_time, False)

        cached_response = self.crud.get_LLMOutputData(
            model=str(self.model),
            parameters=str(self.parameters),
            system_prompt=system_prompt,
            prompt=prompt,
        )

        if cached_response:
            response = json.loads(cached_response)
            cache_used = True
        else:
            response = self.execute_model_call(prompt, system_prompt)
            self._save_response_to_db(prompt, system_prompt, response)
            cache_used = False

        return self._update_response_with_tracking(response, start_time, cache_used)

    def _save_response_to_db(self, prompt, system_prompt, response):
        try:
            output = json.dumps(response)
        except json.JSONDecodeError:
            raise LanguageModelResponseNotJSONError
        self.crud.write_LLMOutputData(
            model=str(self.model),
            parameters=str(self.parameters),
            system_prompt=system_prompt,
            prompt=prompt,
            output=output,
        )

    def get_response(self, prompt: str, system_prompt: str = ""):
        """Get response, parse, and return as string."""
        raw_response = self._get_raw_response(prompt, system_prompt)
        response = self.parse_response(raw_response)
        try:
            dict_response = json.loads(response)
        except json.JSONDecodeError as e:
            print("Could not load JSON. Trying to repair.")
            print(response)
            dict_response, success = repair(response, str(e))
            if not success:
                raise Exception("Even the repair failed.")
        return dict_response

    #######################
    # USEFUL METHODS
    #######################
    def _post_tracker_event(self, raw_response: dict[str, Any]) -> None:
        """Parses the API response and sends usage details to the API Queue."""
        usage = raw_response.get("usage", {})
        usage.update(
            {
                "cached_response": raw_response.get("cached_response", None),
                "elapsed_time": raw_response.get("elapsed_time", None),
                "timestamp": raw_response.get("timestamp", None),
            }
        )
        event = TrackerAPI.APICallDetails(details=usage)
        self.api_queue.put(event)

    def cost(self, raw_response: dict[str, Any]) -> float:
        """Returns the dollar cost of a raw response."""
        keys = raw_response["usage"].keys()
        prices = model_prices.get(self.model)
        return sum([prices.get(key, 0.0) * raw_response["usage"][key] for key in keys])

    #######################
    # FORWARD METHODS
    #######################
    def get_value(self, jobs) -> list[Type[LanguageModel]]:
        """Used by the Jobs class to get the value of the attribute."""
        return jobs.models

    def set_value(self, jobs, new_values: list[Type[LanguageModel]]) -> None:
        """Used by the Jobs class to set the value of the attribute."""
        jobs.models = new_values

    #######################
    # SERIALIZATION METHODS
    #######################
    def to_dict(self) -> dict[str, Any]:
        """Converts instance to a dictionary."""
        return {"model": self.model, "parameters": self.parameters}

    @classmethod
    def from_dict(cls, data: dict) -> Type[LanguageModel]:
        """Converts dictionary to a LanguageModel child instance."""
        from edsl.language_models.registry import get_model_class

        model_class = get_model_class(data["model"])
        data["use_cache"] = True
        return model_class(**data)

    #######################
    # DUNDER METHODS
    #######################
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(model = '{self.model}', parameters={self.parameters})"

    def __add__(self, other_model: Type[LanguageModel]) -> Type[LanguageModel]:
        """Combine two models into a single model (other_model takes precedence over self)"""
        print(
            f"""Warning: one model is replacing another. If you want to run both models, use a single `by` e.g., 
              by(m1, m2, m3) not by(m1).by(m2).by(m3)."""
        )
        return other_model or self



from __future__ import annotations
import random
import textwrap
from typing import Any, Optional, Union
from edsl.questions import Question
from edsl.questions.descriptors import IntegerDescriptor, QuestionOptionsDescriptor
from edsl.scenarios import Scenario
from edsl.utilities import random_string


class QuestionBudget(Question):
    """
    This question asks the respondent to allocate a budget among options.

    Arguments:
    - `budget_sum` is the total amount of the budget to be allocated (positive integer)
    - `question_name` is the name of the question (string)
    - `question_options` are the options the user should allocated the budget to (list of strings)
    - `question_text` is the text of the question (string)

    Optional arguments:
    - `instructions` are the instructions for the question (string). If not provided, the default instructions are used. To view them, run `QuestionBudget.default_instructions`
    - `short_names_dict` maps question_options to short names (dictionary mapping strings to strings)

    To generate an example, run `QuestionBudget.example()`
    """

    question_type = "budget"
    budget_sum: int = IntegerDescriptor(none_allowed=False)
    question_options: list[str] = QuestionOptionsDescriptor()
    default_instructions = textwrap.dedent(
        """\
        You are being asked the following question: {{question_text}}
        The options are 
        {% for option in question_options %}
        {{ loop.index0 }}: {{option}}
        {% endfor %}                       
        Return a valid JSON formatted as follows, with a dictionary for your "answer"
        where the keys are the option numbers and the values are the amounts you want 
        to allocate to the options, and the sum of the values is {{budget_sum}}:
        {"answer": {<put dict of option numbers and allocation amounts here>},
        "comment": "<put explanation here>"}
        Example response for a budget of 100 and 4 options: 
        {"answer": {"0": 25, "1": 25, "2": 25, "3": 25},
        "comment": "I allocated 25 to each option."}
        There must be an allocation listed for each item (including 0).
        """
    )

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_options: list[str],
        budget_sum: int,
        short_names_dict: Optional[dict[str, str]] = None,
        instructions: Optional[str] = None,
    ):
        self.question_name = question_name
        self.question_text = question_text
        self.question_options = question_options
        self.budget_sum = budget_sum
        self.short_names_dict = short_names_dict or {}
        self.instructions = instructions or self.default_instructions

    ################
    # Answer methods
    ################
    def validate_answer(self, answer: dict[str, Any]) -> dict[str, Union[int, str]]:
        self.validate_answer_template_basic(answer)
        self.validate_answer_key_value(answer, "answer", dict)
        self.validate_answer_budget(answer)
        return answer

    def translate_answer_code_to_answer(
        self, answer_codes: dict[str, int], scenario: Scenario = None
    ):
        """Translates the answer codes to the actual answers.
        For example, for a budget question with options ["a", "b", "c"],
        the answer codes are 0, 1, and 2. The LLM will respond with 0.
        This code will translate that to "a".
        """
        translated_codes = []
        for answer_code, response in answer_codes.items():
            translated_codes.append({self.question_options[int(answer_code)]: response})

        return translated_codes

    def simulate_answer(self, human_readable=True):
        "Simulates a valid answer for debugging purposes (what the validator expects)"
        if human_readable:
            keys = self.question_options
        else:
            keys = range(len(self.question_options))
        remaining_budget = self.budget_sum
        values = []
        for _ in range(len(self.question_options)):
            if _ == len(self.question_options) - 1:
                # Assign remaining budget to the last value
                values.append(remaining_budget)
            else:
                # Generate a random value between 0 and remaining budget
                value = random.randint(0, remaining_budget)
                values.append(value)
                remaining_budget -= value
        answer = dict(zip(keys, values))
        return {
            "answer": answer,
            "comment": random_string(),
        }

    ################
    # Helpful methods
    ################
    @classmethod
    def example(cls) -> QuestionBudget:
        return cls(
            question_name="food_budget",
            question_text="How would you allocate $100?",
            question_options=["Pizza", "Ice Cream", "Burgers", "Salad"],
            budget_sum=100,
        )


def main():
    from edsl.questions.QuestionBudget import QuestionBudget

    q = QuestionBudget.example()
    q.question_text
    q.question_options
    q.question_name
    q.short_names_dict
    q.instructions
    # validate an answer
    q.validate_answer(
        {"answer": {0: 100, 1: 0, 2: 0, 3: 0}, "comment": "I like custard"}
    )
    # translate answer code
    q.translate_answer_code_to_answer({0: 100, 1: 0, 2: 0, 3: 0})
    # simulate answer
    q.simulate_answer()
    q.simulate_answer(human_readable=False)
    q.validate_answer(q.simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q



from edsl.exceptions import QuestionSerializationError

# all question types must be registered here
# the key is the question type
# the value is a tuple of the module name and the class name
CLASS_REGISTRY = {
    "budget": ("edsl.questions.QuestionBudget", "QuestionBudget"),
    "checkbox": ("edsl.questions.QuestionCheckBox", "QuestionCheckBox"),
    "extract": ("edsl.questions.QuestionExtract", "QuestionExtract"),
    "free_text": ("edsl.questions.QuestionFreeText", "QuestionFreeText"),
    "functional": ("edsl.questions.QuestionFunctional", "QuestionFunctional"),
    "likert_five": ("edsl.questions.derived.QuestionLikertFive", "QuestionLikertFive"),
    "linear_scale": (
        "edsl.questions.derived.QuestionLinearScale",
        "QuestionLinearScale",
    ),
    "list": ("edsl.questions.QuestionList", "QuestionList"),
    "multiple_choice": (
        "edsl.questions.QuestionMultipleChoice",
        "QuestionMultipleChoice",
    ),
    "numerical": ("edsl.questions.QuestionNumerical", "QuestionNumerical"),
    "rank": ("edsl.questions.QuestionRank", "QuestionRank"),
    "top_k": ("edsl.questions.derived.QuestionTopK", "QuestionTopK"),
    "yes_no": ("edsl.questions.derived.QuestionYesNo", "QuestionYesNo"),
}

question_purpose = {
    "multiple_choice": "When options are known and limited",
    "free_text": "When options are unknown or unlimited",
    "checkbox": "When multiple options can be selected",
    "numerical": "When the answer is a single numerical value e.g., a float",
    "linear_scale": "When options are text, but can be ordered e.g., daily, weekly, monthly, etc.",
    "yes_no": "When the question can be fully answered with either a yes or a no",
}


def get_question_class(question_type: str) -> type:
    """Get the question class for a given question type."""
    if question_type not in CLASS_REGISTRY:
        raise QuestionSerializationError(
            f"No question class registered for type: {question_type}, causing `from_dict` to fail."
        )
    module_name, class_name = CLASS_REGISTRY[question_type]
    question_module = __import__(module_name, globals(), locals(), [class_name], 0)
    return getattr(question_module, class_name)



from contextlib import contextmanager
from sqlalchemy import create_engine
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import sessionmaker, scoped_session
from edsl.config import Config, CONFIG
from edsl.data import Base
from edsl.exceptions import DatabaseConnectionError


class Database:
    """The Database class manages the connection to the database."""

    def __init__(self, config: Config = CONFIG):
        """Initializes the database connection."""
        database_path = config.get("EDSL_DATABASE_PATH")
        try:
            self.engine = create_engine(database_path)
            with self.engine.connect() as _:
                pass
            Base.metadata.create_all(bind=self.engine)
        except SQLAlchemyError as e:
            raise DatabaseConnectionError(str(e)) from None
        self.SessionLocal = scoped_session(
            sessionmaker(autocommit=False, autoflush=False, bind=self.engine)
        )

    @contextmanager
    def get_db(self):
        """Generator that yields a database session."""
        db = None
        try:
            db = self.SessionLocal()
            yield db
        except SQLAlchemyError as e:
            raise DatabaseConnectionError(str(e)) from None
        finally:
            if db is not None:
                self.SessionLocal.remove()


# Note: Python modules are singletons. As such, once `database` is imported
# the same instance of it is reused across the application.
database = Database()



from __future__ import annotations
import random
import textwrap
from jinja2 import Template
from typing import Optional, Union
from edsl.utilities import random_string
from edsl.questions.descriptors import QuestionOptionsDescriptor
from edsl.questions.Question import Question
from edsl.scenarios import Scenario


class QuestionMultipleChoice(Question):
    """
    This question asks the user to select one option from a list of options.

    Arguments:
    - `question_name` is the name of the question (string)
    - `question_options` are the options the user should select from (list of strings)
    - `question_text` is the text of the question (string)

    Optional arguments:
    - `instructions` are the instructions for the question (string). If not provided, the default instructions are used. To view them, run `QuestionMultipleChoice.default_instructions`
    - `short_names_dict` maps question_options to short names (dictionary mapping strings to strings)

    For an example, run `QuestionMultipleChoice.example()`
    """

    question_type = "multiple_choice"
    question_options: list[str] = QuestionOptionsDescriptor()
    default_instructions = textwrap.dedent(
        """\
        You are being asked the following question: {{question_text}}
        The options are 
        {% for option in question_options %}
        {{ loop.index0 }}: {{option}}
        {% endfor %}                       
        Return a valid JSON formatted like this, selecting only the number of the option: 
        {"answer": <put answer code here>, "comment": "<put explanation here>"}
        Only 1 option may be selected.
        """
    )

    def __init__(
        self,
        question_text: str,
        question_options: list[str],
        question_name: str,
        short_names_dict: Optional[dict[str, str]] = None,
        instructions: Optional[str] = None,
    ):
        self.question_text = question_text
        self.question_options = question_options
        self.question_name = question_name
        self.short_names_dict = short_names_dict or dict()
        self.instructions = instructions or self.default_instructions

    ################
    # Answer methods
    ################
    def validate_answer(
        self, answer: dict[str, Union[str, int]]
    ) -> dict[str, Union[str, int]]:
        """Validates the answer"""
        self.validate_answer_template_basic(answer)
        self.validate_answer_multiple_choice(answer)
        return answer

    def translate_answer_code_to_answer(self, answer_code, scenario: Scenario = None):
        """Translates the answer code to the actual answer."""
        scenario = scenario or Scenario()
        translated_options = [
            Template(str(option)).render(scenario) for option in self.question_options
        ]
        return translated_options[int(answer_code)]

    def simulate_answer(
        self, human_readable: bool = True
    ) -> dict[str, Union[int, str]]:
        """Simulates a valid answer for debugging purposes"""
        if human_readable:
            answer = random.choice(self.question_options)
        else:
            answer = random.choice(range(len(self.question_options)))
        return {
            "answer": answer,
            "comment": random_string(),
        }

    ################
    # Example
    ################
    @classmethod
    def example(cls) -> QuestionMultipleChoice:
        return cls(
            question_text="How are you?",
            question_options=["Good", "Great", "OK", "Bad"],
            question_name="how_feeling",
            short_names_dict={"Good": "g", "Great": "gr", "OK": "ok", "Bad": "b"},
        )


def main():
    from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice

    q = QuestionMultipleChoice.example()
    q.question_text
    q.question_options
    q.question_name
    q.short_names_dict
    q.instructions
    # validate an answer
    q.validate_answer({"answer": 0, "comment": "I like custard"})
    # translate answer code
    q.translate_answer_code_to_answer(0, {})
    # simulate answer
    q.simulate_answer()
    q.simulate_answer(human_readable=False)
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q



from __future__ import annotations
from typing import Optional
from edsl.exceptions import QuestionCreationValidationError
from edsl.questions.QuestionCheckBox import QuestionCheckBox


class QuestionTopK(QuestionCheckBox):
    """
    This question asks the user to select exactly K options from a list.

    Arguments:
    - `question_name` is the name of the question (string)
    - `question_text` is the text of the question (string)
    - `question_options` are the options the user should select from (list of strings)
    - `min_selections` is the minimum number of options that must be selected (positive integer).
    - `max_selections` is the maximum number of options that must be selected (positive integer). Must be equal to `min_selections`

    Optional arguments:
    - `instructions` are the instructions for the question (string). If not provided, the default instructions are used. To view them, run `QuestionTopK.default_instructions`
    - `short_names_dict` maps question_options to short names (dictionary mapping strings to strings)

    For an example, run `QuestionTopK.example()`
    """

    question_type = "top_k"

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_options: list[str],
        min_selections: int,
        max_selections: int,
        short_names_dict: Optional[dict[str, str]] = None,
        instructions: Optional[str] = None,
    ):
        super().__init__(
            question_name=question_name,
            question_text=question_text,
            question_options=question_options,
            short_names_dict=short_names_dict,
            min_selections=min_selections,
            max_selections=max_selections,
            instructions=instructions,
        )
        if min_selections != max_selections:
            raise QuestionCreationValidationError(
                "TopK questions must have min_selections == max_selections"
            )
        if min_selections < 1:
            raise QuestionCreationValidationError(
                "TopK questions must have min_selections > 0"
            )

    ################
    # Helpful
    ################
    @classmethod
    def example(cls) -> QuestionTopK:
        return cls(
            question_name="two_fruits",
            question_text="Which of the following fruits do you prefer?",
            question_options=["apple", "banana", "carrot", "durian"],
            min_selections=2,
            max_selections=2,
        )


def main():
    from edsl.questions.derived.QuestionTopK import QuestionTopK

    q = QuestionTopK.example()
    q.question_text
    q.question_options
    q.question_name
    q.short_names_dict
    q.instructions
    # validate an answer
    q.validate_answer({"answer": [0, 3], "comment": "I like custard"})
    # translate answer code
    q.translate_answer_code_to_answer([0, 3], {})
    # simulate answer
    q.simulate_answer()
    q.simulate_answer(human_readable=False)
    q.validate_answer(q.simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q



from __future__ import annotations
from tenacity import retry, wait_exponential, stop_after_attempt
from typing import Any, Type
from edsl.agents import Agent
from edsl.language_models import LanguageModel
from edsl.questions import Question
from edsl.scenarios import Scenario
from edsl.surveys import Survey


class Interview:
    """
    A class that has an Agent answer Survey Questions with a particular Scenario and using a LanguageModel.
    - `Agent.answer_question(question, scenario, model)` is called for each question in the Survey to get an answer to a question.
    - `Survey.gen_path_through_survey()` is called to get a generator that traverses through the Survey.
    - `conduct_interview()` is called to conduct the interview, and returns the answers and comments to the questions in the Survey.
    """

    def __init__(
        self,
        agent: Agent,
        survey: Survey,
        scenario: Scenario,
        model: Type[LanguageModel],
        verbose: bool = False,
        debug: bool = False,
    ):
        self.agent = agent
        self.survey = survey
        self.scenario = scenario
        self.model = model
        self.debug = debug
        self.verbose = verbose
        self.answers: dict[str, str] = {}

    def conduct_interview(
        self, debug: bool = False, replace_missing: bool = True, threaded: bool = False
    ) -> dict[str, Any]:
        """
        Conducts the interview using a generator from the Survey object to traverse through the survey. Sends answers to the Survey, which then sends back the next Question.

        Arguments:
        - `debug`: if True, the agent simulates answers without API calls
        - `replace_missing`: if True, missing answers are imputed with None.

        Returns:
        - `answers`: a dictionary of answers to the survey questions, with keys as question names and values as answers. If the agent also produced a comment, the comment is stored in a key with the question name plus "_comment" appended to it.
        """
        # get the first question
        path_through_survey = self.survey.gen_path_through_survey()
        question = next(path_through_survey)
        survey_inprogress = True

        while survey_inprogress:
            # get agent's response to the question
            if threaded:
                raise NotImplementedError
            else:
                response = self.get_response(question, debug=debug)
            # resolve
            answer = response.get("answer")
            comment = response.pop("comment", None)
            # record the answer
            self.answers[question.question_name] = answer
            if comment:
                self.answers[question.question_name + "_comment"] = comment
            # send the answer to the survey, to get the next question
            try:
                question = path_through_survey.send({question.question_name: answer})
            except StopIteration:
                survey_inprogress = False

        if replace_missing:
            for question_name in self.survey.question_names:
                if question_name not in self.answers:
                    self.answers[question_name] = None

        return self.answers

    @retry(
        wait=wait_exponential(multiplier=2, max=32),
        stop=stop_after_attempt(5),
        reraise=True,
    )
    def get_response(self, question: Question, debug: bool = False) -> dict[str, Any]:
        """Gets the agent's response to a question with exponential backoff."""
        response = self.agent.answer_question(
            question=question, scenario=self.scenario, model=self.model, debug=debug
        )
        return response

    #######################
    # Dunder methods
    #######################
    def __repr__(self) -> str:
        """Returns a string representation of the Interview instance."""
        return f"Interview(agent = {self.agent}, survey = {self.survey}, scenario = {self.scenario}, model = {self.model})"


def main():
    from edsl.language_models import LanguageModelOpenAIThreeFiveTurbo
    from edsl.agents import Agent
    from edsl.surveys import Survey
    from edsl.scenarios import Scenario
    from edsl.questions import QuestionMultipleChoice
    from edsl.jobs.Interview import Interview

    #  a survey with skip logic
    q0 = QuestionMultipleChoice(
        question_text="Do you like school?",
        question_options=["yes", "no"],
        question_name="q0",
    )
    q1 = QuestionMultipleChoice(
        question_text="Why not?",
        question_options=["killer bees in cafeteria", "other"],
        question_name="q1",
    )
    q2 = QuestionMultipleChoice(
        question_text="Why?",
        question_options=["**lack*** of killer bees in cafeteria", "other"],
        question_name="q2",
    )
    s = Survey(questions=[q0, q1, q2])
    s = s.add_rule(q0, "q0 == 'yes'", q2)

    # create an interview
    a = Agent(traits=None)
    scenario = Scenario()
    m = LanguageModelOpenAIThreeFiveTurbo(use_cache=True)
    I = Interview(agent=a, survey=s, scenario=scenario, model=m)

    # conduct five interviews
    for _ in range(5):
        I.conduct_interview(debug=True)

    # replace missing answers
    I
    repr(I)
    eval(repr(I))



from abc import ABC, abstractmethod
import json
from edsl.exceptions import AgentRespondedWithBadJSONError


class InvigilatorBase(ABC):
    def __init__(self, agent, question, scenario, model):
        self.agent = agent
        self.question = question
        self.scenario = scenario
        self.model = model

    @abstractmethod
    def answer_question(self):
        pass


class InvigilatorDebug(InvigilatorBase):
    def answer_question(self):
        return self.question.simulate_answer(human_readable=True)


class InvigilatorHuman(InvigilatorBase):
    def answer_question(self):
        answer = self.agent.answer_question_directly(self.question.question_name)
        response = {"answer": answer}
        response = self.question.validate_response(response)
        response["model"] = "human"
        response["scenario"] = self.scenario
        return response


class InvigilatorFunctional(InvigilatorBase):
    def answer_question(self):
        func = self.question.answer_question_directly
        return func(scenario=self.scenario, agent_traits=self.agent.traits)


class InvigilatorAI(InvigilatorBase):
    def construct_system_prompt(self) -> str:
        """Constructs the system prompt for the LLM call."""
        instruction = self.agent.instruction
        traits = f"Your traits are: {self.agent.traits}."
        return f"{instruction} {traits}"

    def get_response(self, prompt: str, system_prompt):
        """Calls the LLM and gets a response. Used in the `answer_question` method."""
        try:
            response = self.model.get_response(prompt, system_prompt)
        except json.JSONDecodeError as e:
            raise AgentRespondedWithBadJSONError(
                f"Returned bad JSON: {e}"
                f"Prompt: {prompt}"
                f"System Prompt: {system_prompt}"
            )

        return response

    def answer_question(self):
        # actual answers (w/ API call)
        #  get answer
        system_prompt = self.construct_system_prompt()
        prompt = self.question.get_prompt(self.scenario)
        response = self.get_response(prompt, system_prompt)
        #  validate answer
        response = self.question.validate_response(response)
        response = self.question.validate_answer(response)
        answer_code = response["answer"]
        response["answer"] = self.question.translate_answer_code_to_answer(
            answer_code, self.scenario
        )
        return response



import importlib

DEFAULT_MODEL_CLASS = "edsl.language_models.LanguageModelOpenAIFour"

# TODO: Use a meta-class to register models
model_names_to_classes = {
    "gpt-3.5-turbo": "edsl.language_models.LanguageModelOpenAIThreeFiveTurbo",
    "gpt-4": "edsl.language_models.LanguageModelOpenAIFour",
}


def get_model_class(model_name):
    "Returns the class for a given model name"
    class_path = model_names_to_classes.get(model_name, DEFAULT_MODEL_CLASS)
    module_name, class_name = class_path.rsplit(".", 1)
    module = importlib.import_module(module_name)
    cls = getattr(module, class_name)
    return cls


if __name__ == "__main__":
    model_class = get_model_class("gpt-4")
    model = model_class()
    results = model.execute_model_call("Hello world")
    print(results)



from edsl.jobs.JobsRunnerSerial import JobsRunnerSerial
from edsl.jobs.JobsRunnerThreaded import JobsRunnerThreaded
from edsl.jobs.JobsRunnerStreaming import JobsRunnerStreaming

JobsRunnersRegistry = {
    "serial": JobsRunnerSerial,
    "streaming": JobsRunnerStreaming,
    "threaded": JobsRunnerThreaded,
}



from __future__ import annotations
import re
from rich import print
from dataclasses import dataclass

from typing import Any, Generator, Optional, Union
from edsl.exceptions import SurveyCreationError, SurveyHasNoRulesError
from edsl.questions.Question import Question
from edsl.surveys.base import RulePriority, EndOfSurvey
from edsl.surveys.Rule import Rule, RuleCollection

from edsl.Base import Base
from edsl.surveys.SurveyExportMixin import SurveyExportMixin

## Proposed Steps
# 1) Remove name; create a Survey meta-data class
# 2) Remove question_names parameter, as we can always get the names from the questions

from edsl.surveys.descriptors import QuestionsDescriptor


@dataclass
class SurveyMetaData:
    name: str = None
    description: str = None
    version: str = None


class Survey(SurveyExportMixin, Base):
    questions = QuestionsDescriptor()

    """
    A collection of questions that supports skip logic.

    Initalization:
    - `questions`: the questions in the survey (optional)
    - `question_names`: the names of the questions (optional)
    - `name`: the name of the survey (optional)

    Methods:
    -

    Notes:
    - The presumed order of the survey is the order in which questions are added.
    """

    def __init__(
        self,
        questions: list[Question] = None,
        question_names: list[str] = None,
        name: str = None,
        description: str = None,
        version: str = None,
    ):
        """Creates a new survey."""
        self.rule_collection = RuleCollection()
        self.meta_data = SurveyMetaData(
            name=name, description=description, version=version
        )
        self.questions = questions or []

        if question_names is not None:
            print(
                "WARNING: Passing question_names is deprecated. Please use questions with names instead."
            )

    @property
    def name(self):
        print("WARNING: name is deprecated. Please use meta_data.name instead.")
        return self.meta_data.name

    def get_question(self, question_name) -> Question:
        """Returns the question object given the question name"""
        if question_name not in self.question_name_to_index:
            raise KeyError(f"Question name {question_name} not found in survey.")
        index = self.question_name_to_index[question_name]
        return self._questions[index]

    @property
    def question_names(self) -> list[str]:
        """Returns a list of question names in the survey"""
        # return list(self.question_name_to_index.keys())
        return [q.question_name for q in self.questions]

    @property
    def question_name_to_index(self) -> dict[str, int]:
        """Returns a dictionary mapping question names to question indices"""
        return {q.question_name: i for i, q in enumerate(self.questions)}

    def add_question(
        self, question: Question, question_name: Optional[str] = None
    ) -> Survey:
        """
        Adds a question to survey.
        - The question is appended at the end of the self.questions list
        - A default rule is created that the next index is the next question.
        """
        if question_name is not None:
            print(
                "Warning: question_name is deprecated. Please use question.question_name instead."
            )

        if question.question_name in self.question_names:
            raise SurveyCreationError(
                f"""Question name already exists in survey. Please use a different name for the offensing question.
                The problemetic question name is {question_name}.
                """
            )
        index = len(self.questions)
        # TODO: This is a bit ugly because the user
        # doesn't "know" about _questions - it's generated by the
        # descriptor.
        self._questions.append(question)

        # using index + 1 presumes there is a next question
        self.rule_collection.add_rule(
            Rule(
                current_q=index,
                expression="True",
                next_q=index + 1,
                question_name_to_index=self.question_name_to_index,
                priority=RulePriority.DEFAULT.value,
            )
        )
        return self

    def add_stop_rule(self, question: Question, expression: str) -> Survey:
        """Adds a rule that stops the survey."""
        self.add_rule(question, expression, EndOfSurvey())
        return self

    def add_rule(
        self, question: Question, expression: str, next_question: Question
    ) -> Survey:
        """
        Adds a rule to a Question of the Survey with the appropriate priority.
        - The last rule added for the question will have the highest priority.
        - If there are no rules, the rule added gets priority -1.
        """

        def get_question_index(q):
            "Returns the index of the question or EndOfSurvey object"
            if isinstance(q, str):
                question_name = q
                return self.question_name_to_index[question_name]
            elif isinstance(q, Question):
                return self.question_name_to_index[q.question_name]
            elif isinstance(q, EndOfSurvey):
                return EndOfSurvey()
            else:
                raise ValueError(f"Invalid type for question: {type(q)}")

        question_index = get_question_index(question)
        next_question_index = get_question_index(next_question)

        def get_new_rule_priority(question_index):
            priorities = [
                rule.priority
                for rule in self.rule_collection.which_rules(question_index)
            ]
            if len(priorities) == 0:
                priority = RulePriority.DEFAULT.value
            else:
                priority = (
                    max(priorities) + 1
                )  # newer rules take priority over older rules
            return priority

        self.rule_collection.add_rule(
            Rule(
                current_q=question_index,
                expression=expression,
                next_q=next_question_index,
                question_name_to_index=self.question_name_to_index,
                priority=get_new_rule_priority(question_index),
            )
        )

        return self

    ###################
    # FORWARD METHODS
    ###################
    def by(self, *args: Union[Agent, Scenario, LanguageModel]) -> Jobs:
        """Adds Agents, Scenarios, and LanguageModels to a survey and returns a runnable Jobs object"""
        from edsl.jobs.Jobs import Jobs

        job = Jobs(survey=self)
        return job.by(*args)

    def run(self, *args, **kwargs) -> Jobs:
        "Turns the survey into a Job and runs it"
        from edsl.jobs.Jobs import Jobs

        return Jobs(survey=self).run(*args, **kwargs)

    def next_question(
        self, current_question_name: str = None, answers=None
    ) -> Union[Question, EndOfSurvey]:
        """
        Returns the next question in a survey.
        - If called with no arguments, it returns the first question in the survey.
        - If no answers are provided for a question with a rule, the next question is returned. If answers are provided, the next question is determined by the rules and the answers.
        - If the next question is the last question in the survey, an EndOfSurvey object is returned.
        """
        if current_question_name is None:
            return self._questions[0]

        answers = answers or {}
        # print(self.question_name_to_index)
        question_index = self.question_name_to_index[current_question_name]
        next_question_object = self.rule_collection.next_question(
            question_index, answers
        )

        if next_question_object.num_rules_found == 0:
            raise SurveyHasNoRulesError

        if isinstance(next_question_object.next_q, EndOfSurvey):
            return EndOfSurvey()
        else:
            if next_question_object.next_q >= len(self._questions):
                return EndOfSurvey()
            else:
                return self._questions[next_question_object.next_q]

    def gen_path_through_survey(self) -> Generator[Question, dict, None]:
        """
        Generates a coroutine that can be used to conduct an Interview.
        - The coroutine is a generator that yields a question and receives answers.
        - The coroutine starts with the first question in the survey.
        - The coroutine ends when an EndOfSurvey object is returned.
        """
        question = self.next_question()
        question_name = self.question_names[0]
        question.question_name = question_name
        ## We send to the agent the first question and it's name --- say q1 and 'q1'
        ## it sends back a response, which is answers = {'q1': "yes"}
        answers = yield question
        self.answers = answers
        while True:
            question = self.next_question(question.question_name, answers)
            if isinstance(question, EndOfSurvey):
                break
            answers = yield question
            self.answers = answers

    @property
    def scenario_attributes(self) -> list[str]:
        """Returns a list of attributes that admissible Scenarios should have"""
        temp = []
        for question in self._questions:
            question_text = question.question_text
            # extract the contents of all {{ }} in the question text using regex
            matches = re.findall(r"\{\{(.+?)\}\}", question_text)
            # remove whitespace
            matches = [match.strip() for match in matches]
            # add them to the temp list
            temp.extend(matches)
        return temp

    ###################
    # DUNDER METHODS
    ###################
    def __len__(self) -> int:
        """Returns the number of questions in the survey"""
        return len(self._questions)

    def __getitem__(self, index) -> Question:
        """Returns the question object given the question index"""
        return self._questions[index]

    ###################
    # SERIALIZATION METHODS
    ###################
    def to_dict(self) -> dict[str, Any]:
        """Serializes the Survey object to a dictionary."""
        return {
            "questions": [q.to_dict() for q in self._questions],
            "name": self.name,
            "rule_collection": self.rule_collection.to_dict(),
        }

    @classmethod
    def from_dict(cls, data: dict) -> Survey:
        """Deserializes the dictionary back to a Survey object."""
        questions = [Question.from_dict(q_dict) for q_dict in data["questions"]]
        survey = cls(questions=questions, name=data["name"])
        survey.rule_collection = RuleCollection.from_dict(data["rule_collection"])
        return survey

    ###################
    # DISPLAY METHODS
    ###################
    def __repr__(self) -> str:
        """Returns a string representation of the survey"""
        questions_string = ", ".join([repr(q) for q in self._questions])
        question_names_string = ", ".join([repr(name) for name in self.question_names])
        return f"Survey(questions=[{questions_string}], question_names=[{question_names_string}], name={repr(self.name)})"

    def _repr_html_(self) -> str:
        return self.html()

    def show_rules(self) -> None:
        "Prints out the rules in the survey"
        self.rule_collection.show_rules()

    def print(self) -> None:
        "Prints out the survey"
        self.show_questions()

    def show_questions(self):
        "Prints out the questions in the survey"
        for name, question in zip(self.question_names, self._questions):
            print(f"Question:{name},{question}")

    def codebook(self) -> dict[str, str]:
        "Creates a codebook for the survey, mapping question names to question text"
        codebook = {}
        for question in self._questions:
            codebook[question.question_name] = question.question_text
        return codebook

    @classmethod
    def example(cls) -> Survey:
        from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice

        # from edsl.surveys.Survey import Survey

        q0 = QuestionMultipleChoice(
            question_text="Do you like school?",
            question_options=["yes", "no"],
            question_name="q0",
        )
        q1 = QuestionMultipleChoice(
            question_text="Why not?",
            question_options=["killer bees in cafeteria", "other"],
            question_name="q1",
        )
        q2 = QuestionMultipleChoice(
            question_text="Why?",
            question_options=["**lack*** of killer bees in cafeteria", "other"],
            question_name="q2",
        )
        s = cls(questions=[q0, q1, q2])
        s = s.add_rule(q0, "q0 == 'yes'", q2)
        return s


def main():
    def example_survey():
        from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice
        from edsl.surveys.Survey import Survey

        q0 = QuestionMultipleChoice(
            question_text="Do you like school?",
            question_options=["yes", "no"],
            question_name="q0",
        )
        q1 = QuestionMultipleChoice(
            question_text="Why not?",
            question_options=["killer bees in cafeteria", "other"],
            question_name="q1",
        )
        q2 = QuestionMultipleChoice(
            question_text="Why?",
            question_options=["**lack*** of killer bees in cafeteria", "other"],
            question_name="q2",
        )
        s = Survey(questions=[q0, q1, q2])
        s = s.add_rule(q0, "q0 == 'yes'", q2)
        return s

    s = example_survey()
    survey_dict = s.to_dict()
    s2 = Survey.from_dict(survey_dict)
    results = s2.run()
    print(results)



from __future__ import annotations
from abc import ABC, abstractmethod
from edsl.jobs import Jobs
from edsl.results import Results


class JobsRunner(ABC):
    """ABC for JobRunners, which take in a job, conduct interviews, and return their results."""

    def __init__(self, jobs: Jobs):
        self.jobs = jobs
        # create the interviews here so children can use them
        self.interviews = jobs.interviews()

    @abstractmethod
    def run(
        self,
        n: int = 1,
        debug: bool = False,
        verbose: bool = False,
        progress_bar: bool = True,
    ) -> Results:  # pragma: no cover
        """
        Runs the job: conducts Interviews and returns their results.
        - `n`: how many times to run each interview
        - `debug`: prints debug messages
        - `verbose`: prints messages
        - `progress_bar`: shows a progress bar
        """
        raise NotImplementedError



import asyncio

from edsl.jobs import Jobs
from edsl.results import Results, Result
from edsl.jobs.JobsRunner import JobsRunner


class JobsRunnerMultiprocessing(JobsRunner):
    def __init__(self, jobs: Jobs):
        super().__init__(jobs)

    def run(
        self, n=1, verbose=False, sleep=0, debug=False, progress_bar=False
    ) -> Results:
        """Runs a collection of interviews."""

        async def process_task(interview, i):
            answer = await interview.conduct_interview(debug=debug)
            result = Result(
                survey=interview.survey,
                agent=interview.agent,
                scenario=interview.scenario,
                model=interview.model,
                iteration=i,
                answer=answer,
            )
            return result

        async def main(interviews):
            tasks = [
                process_task(interview, i) for i, interview in enumerate(interviews)
            ]
            results = await asyncio.gather(*tasks)
            return results

        data = asyncio.run(main(self.interviews))
        return Results(survey=self.jobs.survey, data=data)



from __future__ import annotations
from collections.abc import Sequence
from itertools import product
from typing import Union
from edsl import CONFIG
from edsl.agents import Agent
from edsl.exceptions import JobsRunError
from edsl.language_models import LanguageModel, LanguageModelOpenAIThreeFiveTurbo
from edsl.results import Results
from edsl.scenarios import Scenario
from edsl.surveys import Survey
from edsl.jobs.base import JobsRunnersRegistry
from edsl.jobs.Interview import Interview
from edsl.api import JobRunnerAPI, ResultsAPI


class Jobs:
    """
    The Jobs class is a collection of agents, scenarios and models and one survey.

    Methods:
    - `by()`: adds agents, scenarios or models to the job. Its a tricksy little method, be careful.
    - `interviews()`: creates a collection of interviews
    - `run()`: runs a collection of interviews

    """

    def __init__(
        self,
        survey: Survey,
        agents: list[Agent] = None,
        models: list[LanguageModel] = None,
        scenarios: list[Scenario] = None,
    ):
        self.survey = survey
        self.agents = agents or []
        self.models = models or []
        self.scenarios = scenarios or []

    def by(
        self,
        *args: Union[
            Agent,
            Scenario,
            LanguageModel,
            Sequence[Union[Agent, Scenario, LanguageModel]],
        ],
    ) -> Jobs:
        """
        Adds Agents, Scenarios and LanguageModels to a job. If no objects of this type exist in the Jobs instance, it stores the new objects as a list in the corresponding attribute. Otherwise, it combines the new objects with existing objects using the object's `__add__` method.

        Arguments:
        - objects or a sequence (list, tuple, ...) of objects of the same type

        Notes:
        - all objects must implement the 'get_value', 'set_value', and `__add__` methods
        - agents: traits of new agents are combined with traits of existing agents. New and existing agents should not have overlapping traits, and do not increase the # agents in the instance
        - scenarios: traits of new scenarios are combined with traits of old existing. New scenarios will overwrite overlapping traits, and do not increase the number of scenarios in the instance
        - models: new models overwrite old models.
        """
        # if the first argument is a sequence, grab it and ignore other arguments
        if len(args) == 1 and isinstance(args[0], Sequence):
            args = args[0]
        # turn the args tuple or the sequence into a list
        args = list(args)
        # grab the first object
        first_object = args[0]
        # use its `get_value` method to get current objects of this type in the Job instance
        current_objects = first_object.get_value(self)
        # if there are no current objects, just store the new objects
        if not current_objects:
            new_objects = args
        # otherwise, combine the current objects with the new objects
        else:
            new_objects = []
            for current_object in current_objects:
                for new_object in args:
                    # use the __add__ method to combine the objects
                    new_objects.append(current_object + new_object)
        # update the Job instance with the new objects
        first_object.set_value(self, new_objects)
        return self

    def interviews(self) -> list[Interview]:
        """
        Returns a list of Interviews, that will eventually be used by the JobRunner.
        - Returns one Interview for each combination of Agent, Scenario, and LanguageModel.
        - If any of Agents, Scenarios, or LanguageModels are missing, fills in with defaults. Note that this will change the corresponding class attributes.
        """
        self.agents = self.agents or [Agent()]
        self.models = self.models or [LanguageModelOpenAIThreeFiveTurbo(use_cache=True)]
        self.scenarios = self.scenarios or [Scenario()]
        interviews = []
        for agent, scenario, model in product(self.agents, self.scenarios, self.models):
            interview = Interview(
                survey=self.survey, agent=agent, scenario=scenario, model=model
            )
            interviews.append(interview)
        return interviews

    def run(
        self,
        method: str = "serial",
        n: int = 1,
        debug: bool = False,
        verbose: bool = False,
        progress_bar: bool = False,
    ) -> Union[Results, ResultsAPI]:
        """
        Runs the Job: conducts Interviews and returns their results.
        - `method`: "serial" or "threaded", defaults to "serial"
        - `n`: how many times to run each interview
        - `debug`: prints debug messages
        - `verbose`: prints messages
        - `progress_bar`: shows a progress bar
        """
        # local mode
        if CONFIG.get("EMERITUS_API_KEY") == "local":
            if method not in JobsRunnersRegistry:
                raise JobsRunError(
                    f"Incorrect method '{method}' provided to .run(). "
                    f"Valid methods are: {list(JobsRunnersRegistry.keys())}"
                )
            JobRunner = JobsRunnersRegistry.get(method)(jobs=self)
            try:
                results = JobRunner.run(
                    n=n, verbose=verbose, debug=debug, progress_bar=progress_bar
                )
            except Exception as e:
                raise JobsRunError(f"Error running job. Exception: {e}.")
            return results
        # remote mode
        else:
            results = JobRunnerAPI(
                api_key=CONFIG.get("EMERITUS_API_KEY"), job_dict=self.to_dict()
            )
            return results

    #######################
    # Dunder methods
    #######################
    def __repr__(self) -> str:
        """Returns an eval-able string representation of the Jobs instance."""
        return f"Jobs(survey={self.survey}, agents={self.agents}, models={self.models}, scenarios={self.scenarios})"

    def __len__(self) -> int:
        """Returns the number of questions that will be asked while running this job."""
        number_of_questions = (
            len(self.agents or [1])
            * len(self.scenarios or [1])
            * len(self.models or [1])
            * len(self.survey)
        )
        return number_of_questions

    #######################
    # Serialization methods
    #######################
    def to_dict(self) -> dict:
        """Converts the Jobs instance to a dictionary."""
        return {
            "survey": self.survey.to_dict(),
            "agents": [agent.to_dict() for agent in self.agents],
            "models": [model.to_dict() for model in self.models],
            "scenarios": [scenario.to_dict() for scenario in self.scenarios],
        }

    @classmethod
    def from_dict(cls, data: dict) -> Jobs:
        """Creates a Jobs instance from a JSON string."""
        return cls(
            survey=Survey.from_dict(data["survey"]),
            agents=[Agent.from_dict(agent) for agent in data["agents"]],
            models=[LanguageModel.from_dict(model) for model in data["models"]],
            scenarios=[Scenario.from_dict(scenario) for scenario in data["scenarios"]],
        )

    #######################
    # Example methods
    #######################
    @classmethod
    def example(cls) -> Jobs:
        from edsl.questions import QuestionMultipleChoice

        q1 = QuestionMultipleChoice(
            question_text="How are you this {{ period }}?",
            question_options=["Good", "Great", "OK", "Terrible"],
            question_name="how_feeling",
        )
        q2 = QuestionMultipleChoice(
            question_text="How were you feeling yesterday {{ period }}?",
            question_options=["Good", "Great", "OK", "Terrible"],
            question_name="how_feeling_yesterday",
        )
        base_survey = Survey(questions=[q1, q2])

        job = base_survey.by(
            Scenario({"period": "morning"}), Scenario({"period": "afternoon"})
        ).by(Agent({"status": "Super duper unhappy"}), Agent({"status": "Joyful"}))

        return job


def main():
    from edsl.jobs import Jobs

    job = Jobs.example()
    len(job) == 8
    results = job.run(debug=True)
    len(results) == 8
    results



class Settings:
    MAX_ANSWER_LENGTH = 2000
    MAX_EXPRESSION_CONSTRAINT_LENGTH = 1000
    MAX_NUM_OPTIONS = 20
    MIN_NUM_OPTIONS = 2
    MAX_OPTION_LENGTH = 1000
    MAX_QUESTION_LENGTH = 100000



import json


def repair(bad_json, error_message=""):
    from edsl.language_models import LanguageModelOpenAIFour

    m = LanguageModelOpenAIFour()
    results = m.execute_model_call(
        f"""Please repair this bad JSON: {bad_json}."""
        + (f"Parsing error message: {error_message}" if error_message else ""),
        system_prompt="You are a helpful agent. Only return the repaired JSON, nothing else.",
    )
    success = True
    try:
        valid_dict = json.loads(results["choices"][0]["message"]["content"])
    except json.JSONDecodeError:
        success = False
        valid_dict = {}
    return valid_dict, success


if __name__ == "__main__":
    bad_json = """
    {
      'answer': "The problematic phrase in the excerpt is \'typically\'. This word is vague and can lead to different interpretations. An alternative phrasing that would be less problematic is: 
      'On average, how long do you cook scrambled eggs?}
    """
    try:
        json.loads(bad_json)
        print("Loaded")
    except json.JSONDecodeError as e:
        error_message = str(e)
        repaired, success = repair(bad_json, error_message)
        print(f"Repaired: {repaired}")



from __future__ import annotations
import gzip
import json
from collections import UserList, defaultdict
from simpleeval import EvalWithCompoundTypes
from typing import Any, Type, Union
from edsl.exceptions import (
    ResultsBadMutationstringError,
    ResultsColumnNotFoundError,
    ResultsInvalidNameError,
    ResultsMutateError,
)
from edsl.agents import Agent
from edsl.data import CRUD
from edsl.language_models import LanguageModel
from edsl.results.Dataset import Dataset
from edsl.results.Result import Result
from edsl.results.ResultsExportMixin import ResultsExportMixin
from edsl.results.RegressionMixin import RegressionMixin
from edsl.results.ResultsOutputMixin import ResultsOutputMixin
from edsl.results.ResultsFetchMixin import ResultsFetchMixin
from edsl.scenarios import Scenario
from edsl.surveys import Survey
from edsl.utilities import (
    is_gzipped,
    is_valid_variable_name,
    print_public_methods_with_doc,
    shorten_string,
)


class Results(
    UserList, ResultsFetchMixin, ResultsExportMixin, ResultsOutputMixin, RegressionMixin
):
    """
    This class is a UserList of Result objects.
    - It is instantiated with a Survey and a list of Result objects (observations).
    - It can be manipulated in various ways with select, filter, mutate, etc.
    - It also has a list of created_columns, which is a list of columns that have been created with `mutate`
    """

    def __init__(
        self,
        survey: Survey,
        data: list[Result],
        created_columns: list = None,
        job_uuid: str = None,
        total_results: int = None,
    ):
        super().__init__(data)
        self.survey = survey
        self.created_columns = created_columns or []
        self._job_uuid = job_uuid
        self._total_results = total_results

    ######################
    # Streaming methods
    ######################
    def _update_results(self) -> None:
        if self._job_uuid and len(self.data) < self._total_results:
            results = [
                Result(
                    agent=Agent.from_dict(json.loads(r.agent)),
                    scenario=Scenario.from_dict(json.loads(r.scenario)),
                    model=LanguageModel.from_dict(json.loads(r.model)),
                    iteration=1,
                    answer=json.loads(r.answer),
                )
                for r in CRUD.read_results(self._job_uuid)
            ]
            self.data = results

    def __repr__(self) -> str:
        self._update_results()
        if self._job_uuid and len(self.data) < self._total_results:
            print(f"Completeness: {len(self.data)}/{self._total_results}")
        return f"Results(data = {self.data}, survey = {self.survey}, created_columns = {self.created_columns})"

    def to_dict(self) -> dict[str, Any]:
        """Converts the Results object to a dictionary"""
        return {
            "data": [result.to_dict() for result in self.data],
            "survey": self.survey.to_dict(),
            "created_columns": self.created_columns,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Results:
        """Converts a dictionary to a Results object"""
        results = cls(
            survey=Survey.from_dict(data["survey"]),
            data=[Result.from_dict(r) for r in data["data"]],
            created_columns=data.get("created_columns", None),
        )
        return results

    @classmethod
    def load(cls, json_file) -> Results:
        """Converts a JSON file to a Results object."""
        mode = "rb" if is_gzipped(json_file) else "r"
        with open(json_file, mode) as f:
            data = json.load(f)
        return cls.from_dict(data)

    def show_methods(self):
        """Prints public methods of the Results class"""
        print_public_methods_with_doc(self)

    ######################
    ## Convenience methods
    ## & Report methods
    ######################
    @property
    def _key_to_data_type(self) -> dict[str, str]:
        """
        Returns a mapping of keys (how_feeling, status, etc.) to strings representing data types (objects such as Agent, Answer, Model, Scenario, etc.)
        - Uses the key_to_data_type property of the Result class.
        - Includes any columns that the user has created with `mutate`
        """
        d = {}
        for result in self.data:
            d.update(result.key_to_data_type)
        for column in self.created_columns:
            d[column] = "answer"
        return d

    @property
    def _data_type_to_keys(self) -> dict[str, str]:
        """
        Returns a mapping of strings representing data types (objects such as Agent, Answer, Model, Scenario, etc.) to keys (how_feeling, status, etc.)
        - Uses the key_to_data_type property of the Result class.
        - Includes any columns that the user has created with `mutate`
        """
        d = defaultdict(set)
        for result in self.data:
            for key, value in result.key_to_data_type.items():
                d[value] = d[value].union(set({key}))
        for column in self.created_columns:
            d["answer"] = d["answer"].union(set({column}))
        return d

    @property
    def answer_keys(self) -> dict[str, str]:
        """Returns a mapping of answer keys to question text"""
        answer_keys = self._data_type_to_keys["answer"]
        answer_keys = {k for k in answer_keys if "_comment" not in k}
        questions_text = [
            self.survey.get_question(k).question_text for k in answer_keys
        ]
        short_question_text = [shorten_string(q, 80) for q in questions_text]
        return dict(zip(answer_keys, short_question_text))

    @property
    def agents(self) -> list[Agent]:
        return [r.agent for r in self.data]

    @property
    def models(self) -> list[Type[LanguageModel]]:
        return [r.model for r in self.data]

    @property
    def scenarios(self) -> list[Scenario]:
        return [r.scenario for r in self.data]

    @property
    def agent_keys(self) -> set[str]:
        """Returns a set of all of the keys that are in the Agent data"""
        return self._data_type_to_keys["agent"]

    @property
    def model_keys(self) -> set[str]:
        """Returns a set of all of the keys that are in the LanguageModel data"""
        return self._data_type_to_keys["model"]

    @property
    def scenario_keys(self) -> set[str]:
        """Returns a set of all of the keys that are in the Scenario data"""
        return self._data_type_to_keys["scenario"]

    @property
    def question_names(self) -> list[str]:
        """Returns a list of all of the question names"""
        return list(self.survey.question_names)

    @property
    def all_keys(self) -> set[str]:
        """Returns a set of all of the keys that are in the Results"""
        answer_keys = set(self.answer_keys)
        return (
            answer_keys.union(self.agent_keys)
            .union(self.scenario_keys)
            .union(self.model_keys)
        )

    def relevant_columns(self) -> set[str]:
        """Returns all of the columns that are in the results."""
        return set().union(
            *(observation.combined_dict.keys() for observation in self.data)
        )

    def _parse_column(self, column: str) -> tuple[str, str]:
        """
        Parses a column name into a tuple containing a data type and a key.
        - Uses the key_to_data_type property of the Results class.
        - The standard way a column is specified is with a dot-separated string, e.g. _parse_column("agent.status")
        - But you can also specify a single key, e.g. "status", in which case it will look up the data type.
        """
        if "." in column:
            data_type, key = column.split(".")
        else:
            try:
                data_type, key = self._key_to_data_type[column], column
            except KeyError:
                raise ResultsColumnNotFoundError(f"Column {column} not found in data")
        return data_type, key

    def first(self) -> Result:
        """Returns the first observation in the results."""
        return self.data[0]

    def mutate(self, new_var_string: str, functions_dict: dict = None) -> Results:
        """
        Creates a value in the Results object as if has been asked as part of the survey.
        - It splits the new_var_string at the "=" and uses simple_eval
        - The functions dict is...

        Example:
        >>> r = Results.create_example()
        >>> r.mutate('how_feeling_x = how_feeling + "x"').select('how_feeling_x')
        [{'answer.how_feeling_x': ['Badx', 'Badx', 'Greatx', 'Greatx']}]
        """
        # extract the variable name and the expression
        if "=" not in new_var_string:
            raise ResultsBadMutationstringError(
                f"Mutate requires an '=' in the string, but '{new_var_string}' doesn't have one."
            )
        raw_var_name, expression = new_var_string.split("=", 1)
        var_name = raw_var_name.strip()
        if not is_valid_variable_name(var_name):
            raise ResultsInvalidNameError(f"{var_name} is not a valid variable name.")

        # create the evaluator
        functions_dict = functions_dict or {}

        def create_evaluator(result: Result) -> EvalWithCompoundTypes:
            return EvalWithCompoundTypes(
                names=result.combined_dict, functions=functions_dict
            )

        def new_result(old_result: Result, var_name: str) -> Result:
            evaluator = create_evaluator(old_result)
            value = evaluator.eval(expression)
            new_result = old_result.copy()
            new_result["answer"][var_name] = value
            return new_result

        try:
            new_data = [new_result(result, var_name) for result in self.data]
        except Exception as e:
            raise ResultsMutateError(f"Error in mutate. Exception:{e}")

        return Results(
            survey=self.survey,
            data=new_data,
            created_columns=self.created_columns + [var_name],
        )

    def select(self, *columns: Union[str, list[str]]) -> Dataset:
        """
        This selects data from the results and turns it into a format like so:
        >>> results = Results.create_example()
        >>> results.select('how_feeling')
        [{'answer.how_feeling': ['Bad', 'Bad', 'Great', 'Great']}]
        """

        if not columns or columns == ("*",) or columns == (None,):
            columns = ("*.*",)

        if isinstance(columns[0], list):
            columns = tuple(columns[0])

        known_data_types = ["answer", "scenario", "agent", "model"]

        def get_data_types_to_return(parsed_data_type):
            if parsed_data_type == "*":  # they want all of the columns
                return known_data_types
            else:
                if parsed_data_type not in known_data_types:
                    raise Exception(
                        f"Data type {parsed_data_type} not found in data. Did you mean one of {known_data_types}"
                    )
                return [parsed_data_type]

        # we're doing to populate this with the data we want to fetch
        to_fetch = defaultdict(list)

        new_data = []
        items_in_order = []
        # iterate through the passed columns
        for column in columns:
            # a user could pass 'result.how_feeling' or just 'how_feeling'
            parsed_data_type, parsed_key = self._parse_column(column)
            data_types = get_data_types_to_return(parsed_data_type)
            found_once = False  # we need to track this to make sure we found the key at least once

            for data_type in data_types:
                # the keys for that data_type e.g.,# if data_type is 'answer', then the keys are 'how_feeling', 'how_feeling_comment', etc.
                relevant_keys = self._data_type_to_keys[data_type]

                for key in relevant_keys:
                    if key == parsed_key or parsed_key == "*":
                        found_once = True
                        to_fetch[data_type].append(key)
                        items_in_order.append(data_type + "." + key)

            if not found_once:
                raise Exception(f"Key {parsed_key} not found in data.")

        for data_type in to_fetch:
            for key in to_fetch[data_type]:
                entries = self._fetch_list(data_type, key)
                new_data.append({data_type + "." + key: entries})

        def sort_by_key_order(dictionary):
            # Extract the single key from the dictionary
            single_key = next(iter(dictionary))
            # Return the index of this key in the list_of_keys
            return items_in_order.index(single_key)

        sorted(new_data, key=sort_by_key_order)

        return Dataset(new_data)

    def sort_by(self, column, reverse=True) -> Results:
        "Sorts the results by a column"

        data_type, key = self._parse_column(column)

        def to_numeric_if_possible(v):
            try:
                return float(v)
            except:
                return v

        new_data = sorted(
            self.data,
            key=lambda x: to_numeric_if_possible(x.get_value(data_type, key)),
            reverse=reverse,
        )
        return Results(survey=self.survey, data=new_data, created_columns=None)

    def filter(self, expression) -> Results:
        """
        This filters a result based on the expression that is passed in.
        >>> r = Results.create_example()
        >>> r.filter("how_feeling == 'Great'").select('how_feeling')
        [{'answer.how_feeling': ['Great', 'Great']}]
        >>> r.filter("how_feeling == 'Nothing'").select('how_feeling')
        [{'answer.how_feeling': []}]
        """

        def create_evaluator(result):
            return EvalWithCompoundTypes(names=result.combined_dict)

        try:
            new_data = [
                result
                for result in self.data
                if create_evaluator(result).eval(expression)
            ]
        except Exception as e:
            print(f"Exception:{e}")

        return Results(survey=self.survey, data=new_data, created_columns=None)

    @classmethod
    def example(cls, debug: bool = False) -> Results:
        """
        Returns an example Results object
        - debug: if False, uses actual API calls
        """
        from edsl.jobs import Jobs

        job = Jobs.example()
        results = job.run(n=1, debug=debug)
        return results


def main():  # pragma: no cover
    """Calls the OpenAI API credits"""
    from edsl.results.Results import Results

    results = Results.example(debug=True)
    print(results.filter("how_feeling == 'Great'").select("how_feeling"))
    print(results.mutate("how_feeling_x = how_feeling + 'x'").select("how_feeling_x"))



from abc import ABC, abstractmethod


class Base(ABC):
    @abstractmethod
    def example(self):
        pass

    @abstractmethod
    def to_dict():
        pass

    @abstractmethod
    def from_dict():
        pass

    @abstractmethod
    def code():
        pass

    def show_methods(self, show_docstrings=True):
        public_methods_with_docstrings = [
            (method, getattr(self, method).__doc__)
            for method in dir(self)
            if callable(getattr(self, method)) and not method.startswith("_")
        ]
        if show_docstrings:
            for method, documentation in public_methods_with_docstrings:
                print(f"{method}: {documentation}")
        else:
            return [x[0] for x in public_methods_with_docstrings]



from edsl.language_models.schemas import model_prices
from edsl.language_models.LanguageModel import LanguageModel
from edsl.language_models.LanguageModelOpenAIThreeFiveTurbo import (
    LanguageModelOpenAIThreeFiveTurbo,
)
from edsl.language_models.LanguageModelOpenAIFour import LanguageModelOpenAIFour



from __future__ import annotations
import copy
import json
from typing import Any, Callable, Optional, Union, Dict
from abc import ABC, abstractmethod

from edsl.Base import Base

from edsl.exceptions import AgentCombinationError

from edsl.agents.Invigilator import (
    InvigilatorDebug,
    InvigilatorHuman,
    InvigilatorFunctional,
    InvigilatorAI,
)

from edsl.language_models import LanguageModel, LanguageModelOpenAIThreeFiveTurbo
from edsl.scenarios import Scenario
from edsl.utilities import (
    create_valid_var_name,
    dict_to_html,
    print_dict_as_html_table,
    print_dict_with_rich,
)

from edsl.agents.descriptors import (
    TraitsDescriptor,
    CodebookDescriptor,
    InstructionDescriptor,
)


class Agent(Base):
    """An agent answers questions."""

    default_instruction = """You are answering questions as if you were a human. Do not break character."""

    traits = TraitsDescriptor()
    codebook = CodebookDescriptor()
    instruction = InstructionDescriptor()

    def __init__(
        self,
        traits: dict = None,
        codebook: dict = None,
        instruction: str = None,
    ):
        self.traits = traits or dict()
        self.codebook = codebook or dict()
        self.instruction = instruction or self.default_instruction

    def agent_with_valid_trait_names(self) -> Agent:
        """
        DEPRECATED
        """
        print(
            "WARNING: This method is deprecated. We now enforce valid names on Agent creation"
        )
        return self

    def answer_question(
        self,
        question: Question,
        scenario: Optional[Scenario] = None,
        model: Optional[LanguageModel] = None,
        debug: bool = False,
    ):
        """
        This is a function where an agent returns an answer to a particular question.
        However, there are several different ways an agent can answer a question, so the
        actual functionality is delegated to an Invigilator object.
        """
        model = model or LanguageModelOpenAIThreeFiveTurbo(use_cache=True)
        scenario = scenario or Scenario()

        if debug:
            # use the question's simulate_answer method
            invigilator_class = InvigilatorDebug
        elif hasattr(question, "answer_question_directly"):
            # it is a functional question and the answer only depends on the agent's traits & the scenario
            invigilator_class = InvigilatorFunctional
        elif hasattr(self, "answer_question_directly"):
            # this of the case where the agent has a method that can answer the question directly
            # this occurrs when 'answer_question_directly' has been monkey-patched onto the agent
            # which happens when the agent is created from an existing survey
            invigilator_class = InvigilatorHuman
        else:
            # this means an LLM agent will be used. This is the standard case.
            invigilator_class = InvigilatorAI

        invigilator = invigilator_class(self, question, scenario, model)
        response = invigilator.answer_question()
        return response

    ################
    # Dunder Methods
    ################
    def __add__(self, other_agent: Agent = None) -> Agent:
        """
        Combines two agents by joining their traits
        - The agents must not have overlapping traits.
        """
        if other_agent is None:
            return self
        elif common_traits := set(self.traits.keys()) & set(other_agent.traits.keys()):
            raise AgentCombinationError(
                f"The agents have overlapping traits: {common_traits}."
            )
        else:
            new_agent = Agent(traits=copy.deepcopy(self.traits))
            new_agent.traits.update(other_agent.traits)
            return new_agent

    def __eq__(self, other: Agent) -> bool:
        """Checks if two agents are equal. Only checks the traits."""
        return self.data == other.data

    def __repr__(self):
        class_name = self.__class__.__name__
        items = [
            f"{k} = '{v}'" if isinstance(v, str) else f"{k} = {v}"
            for k, v in self.data.items()
            if k != "question_type"
        ]
        return f"{class_name}({', '.join(items)})"

    ################
    # Forward methods
    ################
    def to(self, question_or_survey) -> Jobs:
        """Sends an agent to a question or survey."""
        from edsl.agents.AgentList import AgentList

        a = AgentList([self])
        return a.to(question_or_survey)

    def get_value(self, jobs: Jobs) -> list[Agent]:
        """Get a list of agents from a Jobs object. Used in Jobs.by()"""
        return jobs.agents

    def set_value(self, jobs: Jobs, new_values: list[Agent]) -> None:
        """Set the Jobs.agents attribute to the new values. Used in Jobs.by()"""
        jobs.agents = new_values

    ################
    # SERIALIZATION METHODS
    ################
    @property
    def data(self):
        raw_data = {
            k.replace("_", "", 1): v
            for k, v in self.__dict__.items()
            if k.startswith("_")
        }
        if hasattr(self, "set_instructions"):
            if not self.set_instructions:
                raw_data.pop("instruction")
        if self.codebook == {}:
            raw_data.pop("codebook")
        return raw_data

    def to_dict(self) -> dict[str, Union[dict, bool]]:
        """Serializes to a dictionary."""
        return self.data

    @classmethod
    def from_dict(cls, agent_dict: dict[str, Union[dict, bool]]) -> Agent:
        """Deserializes from a dictionary."""
        return cls(**agent_dict)

    ################
    # DISPLAY Methods
    ################
    def dict_to_html(self) -> str:
        """Returns the agent's traits as an HTML table."""
        return dict_to_html(self.traits)

    def print(self, html: bool = False, show: bool = False) -> Optional[str]:
        """Prints the agent's traits as a table."""
        if html:
            return print_dict_as_html_table(self.traits, show)
        else:
            print_dict_with_rich(self.traits)

    @classmethod
    def example(cls) -> Agent:
        """Returns an example agent."""
        return cls(traits={"age": 22, "hair": "brown", "height": 5.5})

    def code(self) -> str:
        """Returns the code for the agent."""
        return f"Agent(traits={self.traits})"


def main():
    """Consumes API credits"""
    from edsl.agents import Agent
    from edsl.questions import QuestionMultipleChoice

    # a simple agent
    agent = Agent(traits={"age": 10, "hair": "brown", "height": 5.5})
    agent.traits
    agent.print()
    # combining two agents
    agent = Agent(traits={"age": 10}) + Agent(traits={"height": 5.5})
    agent.traits
    # Agent -> Job using the to() method
    agent = Agent(traits={"allergies": "peanut"})
    question = QuestionMultipleChoice(
        question_text="Would you enjoy a PB&J?",
        question_options=["Yes", "No"],
        question_name="food_preference",
    )
    job = agent.to(question)
    # run the job
    results = job.run()
    # results


if __name__ == "__main__":
    main()



from __future__ import annotations
import textwrap
from typing import Any, Optional
from edsl.questions import Question
from edsl.questions.descriptors import QuestionAllowNonresponseDescriptor
from edsl.scenarios import Scenario
from edsl.utilities import random_string


# TODO: should allow answer = {"answer": None} if allow_nonresponse is True
class QuestionFreeText(Question):
    """
    This question asks the user to answer with free text.

    Arguments:
    - `question_name` is the name of the question (string)
    - `question_text` is the text of the question (string)

    Optional arguments:
    - `allow_nonresponse` is whether the user can skip the question (boolean). If not provided, the default is False.
    - `instructions` are the instructions for the question (string). If not provided, the default instructions are used. To view them, run `QuestionFreeText.default_instructions`

    For an example, run `QuestionFreeText.example()`
    """

    question_type = "free_text"
    allow_nonresponse: bool = QuestionAllowNonresponseDescriptor()
    default_instructions = textwrap.dedent(
        """\
        You are being asked the following question: {{question_text}}
        Return a valid JSON formatted like this: 
        {"answer": "<put free text answer here>"}
        """
    )

    def __init__(
        self,
        question_text: str,
        question_name: str,
        allow_nonresponse: Optional[bool] = None,
        instructions: Optional[str] = None,
    ):
        self.question_text = question_text
        self.question_name = question_name
        self.allow_nonresponse = allow_nonresponse or False
        self.instructions = instructions or self.default_instructions

    ################
    # Answer methods
    ################
    def validate_answer(self, answer: Any) -> dict[str, str]:
        self.validate_answer_template_basic(answer)
        self.validate_answer_key_value(answer, "answer", str)
        return answer

    def translate_answer_code_to_answer(self, answer, scenario: Scenario = None):
        """Does nothing, because the answer is already in a human-readable format."""
        return answer

    def simulate_answer(self, human_readable: bool = True) -> dict[str, str]:
        return {"answer": random_string()}

    @classmethod
    def example(cls) -> QuestionFreeText:
        return cls(
            question_name="how_are_you",
            question_text="How are you?",
            allow_nonresponse=True,
        )


def main():
    from edsl.questions.QuestionFreeText import QuestionFreeText

    q = QuestionFreeText.example()
    q.question_text
    q.question_name
    q.instructions
    # validate an answer
    q.validate_answer({"answer": "I like custard"})
    # translate answer code
    q.translate_answer_code_to_answer({"answer"})
    # simulate answer
    q.simulate_answer()
    q.simulate_answer(human_readable=False)
    q.validate_answer(q.simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q



from __future__ import annotations
import threading
import time
from threading import Lock, Event
from queue import Queue
from edsl.jobs import Jobs
from edsl.jobs.Worker import Worker
from edsl.trackers.TrackerAPI import TrackerAPI
from edsl.trackers.TrackerTasks import TrackerTasks
from edsl.results import Results
from edsl.jobs.JobsRunner import JobsRunner

"""
This module contains the JobsRunnerThreaded class, which is a threaded version of the JobsRunner class.
It (cautiously) addes more workers to the workforce if the API tracker observes that the API limit is not being reached.
If the API tracker observes that the API limit is being reached, it stops adding workers to the workforce.
"""


class JobsRunnerThreaded(JobsRunner):
    def __init__(self, jobs: Jobs):
        super().__init__(jobs)
        self.task_queue = Queue()
        self.event_queue = Queue()
        self.results_queue = Queue()
        self.api_queue = self.interviews[0].model.api_queue
        self.worker_threads = []
        self.lock = Lock()
        self.number_of_workers = 1
        self.sleep_time_between_workers_seconds = 0.1
        self.adjustment_interval_seconds = 5

    def change_worker_pace_multiplicatively(self, multiplier):
        "Increases the time between workers by some percentage."
        with self.lock:
            self.sleep_between_workers_seconds *= multiplier

    def add_worker(self, tracker_api, tracker_tasks) -> None:
        "Adds a worker to the worker pool."
        worker_instance = Worker(tracker_api=tracker_api, tracker_tasks=tracker_tasks)
        thread = threading.Thread(
            target=worker_instance,
            args=(
                self.task_queue,
                self.results_queue,
                self.event_queue,
            ),
        )
        thread.start()
        self.worker_threads.append(thread)

    def dynamic_labor_adjustment(self, all_done, tracker_api, tracker_tasks) -> None:
        """Adjusts the number of workers based on the API tracker's observations."""
        limit_ever_exceeded = False
        while not all_done.is_set():
            time.sleep(self.adjustment_interval_seconds)
            usage_rates = tracker_api.usage_rates()
            if usage_rates.pct_of_tpm_limit < 30 and not limit_ever_exceeded:
                self.add_worker(tracker_api, tracker_tasks)
            else:
                limit_ever_exceeded = True

    def run(
        self, n=1, verbose=False, sleep=0, debug=False, progress_bar=False
    ) -> Results:
        """Runs a collection of interviews."""
        all_done = Event()  # a thread-safe event that is set when all tasks are done

        # Add tasks (interviews) to the queue
        [self.task_queue.put(interview) for interview in self.interviews]

        # instantiate the tracker, adding it to a thread and joining the thread
        tracker_tasks = TrackerTasks(
            num_interviews=len(self.interviews),
            lock=self.lock,
            monitored_queue=self.event_queue,
        )
        tasks_monitoring_thread = threading.Thread(
            target=tracker_tasks, args=(all_done,)
        )
        tasks_monitoring_thread.start()

        # instantiate the API tracker, adding it to a thread and joining the thread
        tracker_api = TrackerAPI(lock=self.lock, monitored_queue=self.api_queue)
        api_monitoring_thread = threading.Thread(target=tracker_api, args=(all_done,))
        api_monitoring_thread.start()

        workforce_adjust_thread = threading.Thread(
            target=self.dynamic_labor_adjustment,
            args=(all_done, tracker_api, tracker_tasks),
        )
        workforce_adjust_thread.start()

        # instantiate workers, adding them to threads and joining the threads
        [
            self.add_worker(tracker_api, tracker_tasks)
            for _ in range(self.number_of_workers)
        ]

        # Waiting for tasks to be completed
        self.task_queue.join()

        all_done.set()

        # Shutting down workforce adjustment thread.
        workforce_adjust_thread.join()

        # Shutting down tasks monitoring thread.
        tasks_monitoring_thread.join()

        # "Shutting down API monitoring thread."
        api_monitoring_thread.join()

        # Emptying the results queue
        data = []
        while not self.results_queue.empty():
            data.append(self.results_queue.get())

        return Results(survey=self.jobs.survey, data=data)



import random
import time
import threading
from queue import Empty
from rich import print as rprint
from edsl.results.Result import Result
from edsl.utilities.interface import heartbeat_generator
from edsl.trackers.TrackerTasks import TrackerTasks


class Worker:
    """A worker does the task of taking interviews off the queue and conducting them.
    It sends updates to the event queue about its status.

    - A worker is taking events off the task_queue (get)
    - A worker is sending events to the event_queue describing its status (put)
    - A worker is monitoring the api_tracker to see if it is rate limited

    """

    def __init__(self, tracker_api, tracker_tasks, config=None):
        self.config = config
        self.tracker_api = tracker_api
        self.tracker_tasks = tracker_tasks
        self.wait_time_after_rate_limit_error = 60

        self.heartbeat_gen = heartbeat_generator()

    def print_status_update(self):
        usage_rates = self.tracker_api.usage_rates()
        pct_complete = round(self.tracker_tasks.percentage_complete)
        active_workers = self.tracker_tasks.active_workers
        estimated_length = 100
        heartbeat = next(self.heartbeat_gen)
        status_str = (
            f"{heartbeat}; "
            f"{pct_complete}% of tasks complete; "
            f"{active_workers} worker(s); "
            f"Est. TPM (k): {usage_rates.estimated_tokens_per_minute_k: ,} ({usage_rates.pct_of_tpm_limit}% of lim.); "
            f"Est. RPM (k): {usage_rates.estimated_requests_per_minute_k:,} ({usage_rates.pct_of_rpm_limit}% of lim.)); "
        )
        # Pad with spaces to reach the estimated length
        padded_status_str = status_str.ljust(estimated_length)
        print(f"\r{padded_status_str}", end="", flush=True)

    def __call__(self, task_queue, results_queue, event_queue, debug=False):
        event_queue.put(TrackerTasks.WorkerActivated())
        thread_status = TrackerTasks.ThreadStatus(
            thread_id=threading.get_ident(), status="Running"
        )
        event_queue.put(thread_status)
        while True:
            self.print_status_update()
            sleep_time = random.random() * 0.1
            time.sleep(sleep_time)  # prevent thundering herd of synchronization
            try:
                interview = task_queue.get(block=True, timeout=1)
            except Empty as e:
                break

            try:
                event_queue.put(TrackerTasks.TaskStarted())
                answer = interview.conduct_interview(debug=debug)
                result = Result(
                    agent=interview.agent,
                    scenario=interview.scenario,
                    model=interview.model,
                    iteration=0,
                    answer=answer,
                )
                results_queue.put(result)
            except Exception as e:  # get the actual name of the exception
                print(f"Caught an exception of type {type(e).__name__}")
                print(f"Text:{e}")
                rprint("[red]Rate limit error[/red]")
                sleep_amount = round(random.random() * 60)
                print(f"Putting this worker to sleep for {sleep_amount} seconds")
                time.sleep(sleep_amount)
                task_queue.put(interview)
                # raise Exception
            else:
                event_queue.put(TrackerTasks.TaskCompleted())
                task_queue.task_done()

        event_queue.put(TrackerTasks.WorkerActivated())



from __future__ import annotations
import textwrap
from abc import ABC, abstractmethod
from jinja2 import Template, Environment, meta
from typing import Any, Type
from edsl.exceptions import (
    QuestionResponseValidationError,
    QuestionSerializationError,
    QuestionScenarioRenderError,
)
from edsl.questions.descriptors import (
    InstructionsDescriptor,
    QuestionNameDescriptor,
    QuestionTextDescriptor,
    ShortNamesDictDescriptor,
)
from edsl.questions.question_registry import get_question_class
from edsl.questions.AnswerValidatorMixin import AnswerValidatorMixin


class Question(ABC, AnswerValidatorMixin):
    """
    ABC for something.
    """

    question_name: str = QuestionNameDescriptor()
    question_text: str = QuestionTextDescriptor()
    short_names_dict: dict[str, str] = ShortNamesDictDescriptor()
    instructions: str = InstructionsDescriptor()

    @property
    def data(self) -> dict:
        """Returns a dictionary of question attributes **except** for question_type"""
        # question-specific attributes start with an underscore
        candidate_data = {
            k.replace("_", "", 1): v
            for k, v in self.__dict__.items()
            if k.startswith("_")
        }
        # things (?) that over-ride defaults
        optional_attributes = {
            "set_instructions": "instructions",
        }
        # ?
        for boolean_flag, attribute in optional_attributes.items():
            if hasattr(self, boolean_flag) and not getattr(self, boolean_flag):
                candidate_data.pop(attribute, None)

        return candidate_data

    ############################
    # Serialization methods
    ############################
    def to_dict(self) -> dict[str, Any]:
        """Converts the question to a dictionary that includes the question type (useful for deserialization)."""
        candidate_data = self.data.copy()
        candidate_data["question_type"] = self.question_type
        return candidate_data

    @classmethod
    def from_dict(cls, data: dict) -> Type[Question]:
        """Constructs a question object from a dictionary created by that question's `to_dict` method."""
        local_data = data.copy()
        try:
            question_type = local_data.pop("question_type")
        except:
            raise QuestionSerializationError(
                f"Cannot deserialize question data because it does not have a 'question_type' field. "
                f"Data: {data}"
            )
        question_class = get_question_class(question_type)
        return question_class(**local_data)

    ############################
    # Dunder methods
    ############################
    def __repr__(self) -> str:
        """Returns a string representation of the question. Should be able to be used to reconstruct the question."""
        class_name = self.__class__.__name__
        items = [
            f"{k} = '{v}'" if isinstance(v, str) else f"{k} = {v}"
            for k, v in self.data.items()
            if k != "question_type"
        ]
        return f"{class_name}({', '.join(items)})"

    def __eq__(self, other: Type[Question]) -> bool:
        """Checks if two questions are equal. Equality is defined as having the .to_dict()"""
        if not isinstance(other, Question):
            return False
        return self.to_dict() == other.to_dict()

    # TODO: Throws an error that should be addressed at QuestionFunctional
    def __add__(self, other_question):
        """
        Composes two questions into a single question.
        >>> from edsl.scenarios.Scenario import Scenario
        >>> from edsl.questions.QuestionFreeText import QuestionFreeText
        >>> from edsl.questions.QuestionNumerical import QuestionNumerical
        >>> q1 = QuestionFreeText(question_text = "What is the capital of {{country}}", question_name = "capital")
        >>> q2 = QuestionNumerical(question_text = "What is the population of {{capital}}, in millions. Please round", question_name = "population")
        >>> q3 = q1 + q2
        >>> Scenario({"country": "France"}).to(q3).run().select("capital_population")
        ['2']
        """
        from edsl.questions import compose_questions

        return compose_questions(self, other_question)

    ############################
    # LLM methods
    ############################
    @staticmethod
    def scenario_render(text: str, scenario_dict: dict[str, Any]) -> str:
        """
        Replaces the variables in the question text with the values from the scenario.
        - We allow nesting, and hence we may need to do this many times. There is a nesting limit of 100.
        """
        t = text
        MAX_NESTING = 100
        counter = 0
        while True:
            counter += 1
            new_t = Template(t).render(scenario_dict)
            if new_t == t:
                break
            t = new_t
            if counter > MAX_NESTING:
                raise QuestionScenarioRenderError(
                    "Too much nesting - you created an infnite loop here, pal"
                )

        return new_t

    def formulate_prompt(self, traits=None, focal_item=None):
        """
        Builds the prompt to send to the LLM. The system prompt contains:
        - Context that might be helpful
        - The traits of the agent
        - The focal item and a description of what it is.
        """
        system_prompt = ""
        instruction_part = textwrap.dedent(
            """\
        You are answering questions as if you were a human. 
        Do not break character.  
        """
        )
        system_prompt += instruction_part

        if traits is not None:
            relevant_trait = traits.relevant_traits(self)
            traits_part = f"Your traits are: {relevant_trait}"
            system_prompt += traits_part

        prompt = ""
        if focal_item is not None:
            focal_item_prompt_fragment = textwrap.dedent(
                f"""\
            The question you will be asked will be about a {focal_item.meta_description}.
            The particular one you are responding to is: {focal_item.content}.
            """
            )
            prompt += focal_item_prompt_fragment

        prompt += self.get_prompt()
        return prompt, system_prompt

    def get_prompt(self, scenario=None) -> str:
        """Shows which prompt should be used with the LLM for this question.
        It extracts the question attributes from the instantiated question data model.
        """
        scenario = scenario or {}
        template = Template(self.instructions)
        template_with_attributes = template.render(self.data)
        env = Environment()
        ast = env.parse(template_with_attributes)
        undeclared_variables = meta.find_undeclared_variables(ast)
        if any([v not in scenario for v in undeclared_variables]):
            raise QuestionScenarioRenderError(
                f"Scenario is missing variables: {undeclared_variables}"
            )
        prompt = self.scenario_render(template_with_attributes, scenario)
        return prompt

    @abstractmethod
    def validate_answer(self, answer: dict[str, str]):
        pass

    def validate_response(self, response):
        """Validates the response from the LLM"""
        if "answer" not in response:
            raise QuestionResponseValidationError(
                "Response from LLM does not have an answer"
            )
        return response

    @abstractmethod
    def translate_answer_code_to_answer(self):  # pragma: no cover
        """Translates the answer code to the actual answer. Behavior depends on the question type."""
        pass

    @abstractmethod
    def simulate_answer(self, human_readable=True) -> dict:  # pragma: no cover
        """Simulates a valid answer for debugging purposes (what the validator expects)"""
        pass

    ############################
    # Forward methods
    ############################
    def add_question(self, other):
        "Adds a question to this question by turning them into a survey with two questions"
        from edsl.surveys.Survey import Survey

        s = Survey([self, other], [self.question_name, other.question_name])
        return s

    def run(self, *args, **kwargs):
        "Turns a single question into a survey and runs it."
        from edsl.surveys.Survey import Survey

        s = Survey([self], [self.question_name])
        return s.run(*args, **kwargs)

    def by(self, *args):
        "Documentation missing."
        from edsl.surveys.Survey import Survey

        s = Survey([self], [self.question_name])
        return s.by(*args)



from __future__ import annotations
from collections import UserList
from typing import Optional, Union
from edsl.agents import Agent
from edsl.agents.AgentListExportMixin import AgentListExportMixin
from edsl.Base import Base


class AgentList(UserList, Base, AgentListExportMixin):
    def __init__(self, data: Optional[list] = None):
        if data is not None:
            super().__init__(data)
        else:
            super().__init__()

    def to(self, question_or_survey: Union["Question", "Survey"]):
        return question_or_survey.by(*self)

    def update_traits(self, new_attributes: list):
        for agent in self.data:
            agent.update_traits(new_attributes)

    def print(self, html=False):
        html = ""
        for agent in self.data:
            html += agent.dict_to_html()
        from edsl.utilities.interface import gen_html_sandwich
        from edsl.utilities.interface import view_html

        view_html(gen_html_sandwich(html))

    def to_dict(self):
        return {"agent_list": [agent.to_dict() for agent in self.data]}

    @classmethod
    def from_dict(cls, data: dict) -> "AgentList":
        """Deserializes the dictionary back to an Agent List object."""
        agents = [Agent.from_dict(agent_dict) for agent_dict in data["agent_list"]]
        return cls(agents)

    @classmethod
    def example(cls):
        return cls([Agent.example(), Agent.example()])

    def code(self):
        lines = [
            "from edsl.agents.Agent import Agent",
            "from edsl.agents.AgentList import AgentList",
        ]
        lines.append(f"agent_list = AgentList({self.data})")
        return lines


if __name__ == "__main__":
    a = AgentList([1, 2, 3])
    print(a)

    from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice
    from edsl.agents.Agent import Agent

    q = QuestionMultipleChoice(
        question_text="How are you feeling?",
        question_options=["Very sad.", "Sad.", "Neutral.", "Happy.", "Very happy."],
        question_name="feelings",
    )

    a = Agent(traits={"feeling": "Very sad."})
    agent_list = AgentList([a])
    results = agent_list.to(q).run()

    print(results)

    results2 = q.by(agent_list).run()

    print(results2)



import copy
from collections import UserDict

from edsl.Base import Base


class Scenario(UserDict, Base):
    """A Scenario is a dictionary of key/values that describe some situation."""

    def __add__(self, other_scenario):
        """Combine two scenarios. If the other scenario is None, then just return self.
        >>> s1 = Scenario({"price": 100, "quantity": 2})
        >>> s2 = Scenario({"color": "red"})
        >>> s1 + s2
        {'price': 100, 'quantity': 2, 'color': 'red'}
        >>> type(s1 + s2)
        <class '__main__.Scenario'>
        """
        if other_scenario is None:
            return self
        else:
            new_scenario = Scenario()
            new_scenario.data = copy.deepcopy(self.data)
            new_scenario.update(copy.deepcopy(other_scenario))
            return Scenario(new_scenario)

    def to(self, question_or_survey) -> "Jobs":
        """Run a question/survey with this particular scenario.
        Useful if you want to reverse the typical chain of operations.

        >>> from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice
        >>> s = Scenario({"food": "wood chips"})
        >>> q = QuestionMultipleChoice(question_text = "Do you enjoy the taste of {{food}}?", question_options = ["Yes", "No"], question_name = "food_preference")
        >>> s.to(q)
        Jobs(survey=Survey(questions=[QuestionMultipleChoice(question_text = "Do you enjoy the taste of {{food}}?", question_options = ['Yes', 'No'], question_name = "food_preference")], question_names=['food_preference'], name = None), agents=[Agent(traits = {})], models=[LanguageModelOpenAIThreeFiveTurbo(model = "gpt-3.5-turbo", use_cache = True)], scenarios=[{'food': 'wood chips'}])
        """
        return question_or_survey.by(self)

    def rename(self, replacement_dict: dict) -> "Scenario":
        """Rename the keys of a scenario. Useful for changing the names of keys.

        >>> s = Scenario({"food": "wood chips"})
        >>> s.rename({"food": "food_preference"})
        {'food_preference': 'wood chips'}
        """
        new_scenario = Scenario()
        for key, value in self.items():
            if key in replacement_dict:
                new_scenario[replacement_dict[key]] = value
            else:
                new_scenario[key] = value
        return new_scenario

    def make_question(self, question_class: type):
        """Make a question from this scenario. Note it takes a QuestionClass (not a question)
        as an input.

        >>> from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice
        >>> from edsl.agents.Agent import Agent

        >>> s = Scenario({"question_name": "feelings",
        ...               "question_text": "How are you feeling?",
        ...               "question_options": ["Very sad.", "Sad.", "Neutral.", "Happy.", "Very happy."]})
        >>> q = s.make_question(QuestionMultipleChoice)
        >>> q.by(Agent(traits = {'feeling': 'Very sad'})).run().select("feelings")
        ['Very sad.']
        """
        return question_class(**self)

    def get_value(self, jobs):
        """Get the scenarios from an existing Jobs instance."""
        return jobs.scenarios

    def set_value(self, jobs, new_values):
        """Set the scenarios from an existing Jobs instance."""
        jobs.scenarios = new_values

    def to_dict(self):
        """Convert a scenario to a dictionary.
        >>> s = Scenario({"food": "wood chips"})
        >>> s.to_dict()
        {'food': 'wood chips'}
        """
        return self.data

    @classmethod
    def from_dict(cls, d):
        """Convert a dictionary to a scenario.
        >>> Scenario.from_dict({"food": "wood chips"})
        {'food': 'wood chips'}
        """
        return cls(d)

    @classmethod
    def example(cls):
        """Returns an example scenario.
        >>> Scenario.example()
        {'scenario': 'example'}
        """
        return cls(
            {
                "persona": "A reseacher studying whether LLMs can be used to generate surveys."
            }
        )

    def code(self):
        """Returns the code for the scenario."""
        lines = []
        lines.append("from edsl.scenario import Scenario")
        return f"Scenario({self.data})"


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import openai
import re
from typing import Any
from edsl import CONFIG
from edsl.language_models import LanguageModel

openai.api_key = CONFIG.get("OPENAI_API_KEY")


class LanguageModelOpenAIFour(LanguageModel):
    """
    Child class of LanguageModel for interacting with OpenAI GPT-4 model.
    """

    _model_ = "gpt-4-1106-preview"
    # _model_ = "gpt-4"
    _parameters_ = {
        "temperature": 0.5,
        "max_tokens": 1000,
        "top_p": 1,
        "frequency_penalty": 0,
        "presence_penalty": 0,
        "use_cache": True,
    }

    def __init__(self, **kwargs):
        self.model = self._model_
        # set parameters, with kwargs taking precedence over defaults
        self.parameters = dict({})
        for parameter, default_value in self._parameters_.items():
            if parameter in kwargs:
                self.parameters[parameter] = kwargs[parameter]
            else:
                self.parameters[parameter] = default_value
                kwargs[parameter] = default_value
        super().__init__(**kwargs)

    def execute_model_call(
        self, prompt: str, system_prompt: str = ""
    ) -> dict[str, Any]:
        """Calls the OpenAI API and returns the API response."""
        return openai.chat.completions.create(
            model=self.model,
            messages=[
                {"role": "system", "content": system_prompt},
                {"role": "user", "content": prompt},
            ],
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            top_p=self.top_p,
            frequency_penalty=self.frequency_penalty,
            presence_penalty=self.presence_penalty,
        ).model_dump()

    @staticmethod
    def parse_response(raw_response: dict[str, Any]) -> str:
        """Parses the API response and returns the response text."""
        response = raw_response["choices"][0]["message"]["content"]
        pattern = r"^```json(?:\\n|\n)(.+?)(?:\\n|\n)```$"
        match = re.match(pattern, response, re.DOTALL)
        if match:
            return match.group(1)
        else:
            return response


def main():
    from edsl.language_models import LanguageModelOpenAIFour

    m = LanguageModelOpenAIFour(use_cache=False)
    m
    m.execute_model_call(prompt="How are you?")
    m.execute_model_call(
        system_prompt="Pretend you are human. Do not break character. Only respond shortly, without asking any questions.",
        prompt="How are you?",
    )
    raw_english = m.get_raw_response(
        system_prompt="Pretend you are human. Do not break character. Only respond shortly, without asking any questions.",
        prompt="What is your favorite color?",
    )
    print(m.parse_response(raw_english))
    print(m.cost(raw_english))

    # ----
    system_prompt = "Pretend you are human. Do not break character. Only respond shortly, without asking any questions."
    prompt = "What is your favorite color?"
    m = LanguageModelOpenAIFour(use_cache=True)
    # the execute model call should be a dict
    raw_german = m.execute_model_call(system_prompt=system_prompt, prompt=prompt)
    raw_german = m.get_raw_response(system_prompt=system_prompt, prompt=prompt)
    print(raw_german)
    print(m.parse_response(raw_german))
    print(m.cost(raw_german))



from enum import Enum


class RulePriority(Enum):
    DEFAULT = -1


class EndOfSurvey:
    "A named object that represents the end of the survey."
    pass

    def __repr__(self):
        return "EndOfSurvey"

    def __str__(self):
        return "EndOfSurvey"



from __future__ import annotations
from collections import UserDict
from typing import Any, Type
from edsl.agents import Agent
from edsl.language_models import LanguageModel
from edsl.scenarios import Scenario


class Result(UserDict):
    """
    This class captures the result of one interview.
    - Its main data is an Agent, a Scenario, a Model, an Iteration, and an Answer.
    - These are stored both in the UserDict and as attributes.
    """

    def __init__(
        self,
        agent: Agent,
        scenario: Scenario,
        model: Type[LanguageModel],
        iteration: int,
        answer: str,
    ):
        # initialize the UserDict
        data = {
            "agent": agent.agent_with_valid_trait_names(),
            "scenario": scenario,
            "model": model,
            "iteration": iteration,
            "answer": answer,
        }
        super().__init__(**data)
        # but also store the data as attributes
        self.agent = agent.agent_with_valid_trait_names()
        self.scenario = scenario
        self.model = model
        self.iteration = iteration
        self.answer = answer

    ###############
    # Used in Results
    ###############
    @property
    def sub_dicts(self) -> dict[str, dict]:
        """Returns a dictionary where keys are strings for each of the main class attributes/objects (except for iteration) and values are dictionaries for the attributes and values for each of these objects."""
        return {
            "agent": self.agent.agent_with_valid_trait_names().traits,
            "scenario": self.scenario,
            "model": self.model.parameters | {"model": self.model.model},
            "answer": self.answer,
        }

    @property
    def combined_dict(self) -> dict[str, Any]:
        """Returns a dictionary that includes all sub_dicts, but also puts the key-value pairs in each sub_dict as a key_value pair in the combined dictionary."""
        combined = {}
        for key, sub_dict in self.sub_dicts.items():
            combined.update(sub_dict)
            combined.update({key: sub_dict})
        return combined

    def get_value(self, data_type: str, key: str) -> Any:
        """Returns the value for a given data type and key
        - data types can be "agent", "scenario", "model", or "answer"
        - keys are relevant attributes of the Objects the data types represent
        results.get_value("answer", "how_feeling") will return "Good" or "Bad" or whatnot
        """
        return self.sub_dicts[data_type][key]

    @property
    def key_to_data_type(self) -> dict[str, str]:
        """Returns a dictionary where keys are object attributes and values are the data type (object) that the attribute is associated with."""
        d = {}
        for data_type in ["agent", "scenario", "model", "answer"]:
            for key in self.sub_dicts[data_type]:
                d[key] = data_type
        return d

    ###############
    # Useful
    ###############
    def copy(self) -> Result:
        """Returns a copy of the Result object."""
        return Result.from_dict(self.to_dict())

    def __repr__(self):
        return f"Result(agent={self.agent}, scenario={self.scenario}, model={self.model}, iteration={self.iteration}, answer={self.answer})"

    def __eq__(self, other):
        return self.to_dict() == other.to_dict()

    ###############
    # Serialization
    ###############
    def to_dict(self) -> dict[str, Any]:
        """Returns a dictionary representation of the Result object."""
        return {
            k: v if not hasattr(v, "to_dict") else v.to_dict() for k, v in self.items()
        }

    @classmethod
    def from_dict(self, json_dict: dict) -> Result:
        """Returns a Result object from a dictionary representation."""
        result = Result(
            agent=Agent.from_dict(json_dict["agent"]),
            scenario=Scenario.from_dict(json_dict["scenario"]),
            model=LanguageModel.from_dict(json_dict["model"]),
            iteration=json_dict["iteration"],
            answer=json_dict["answer"],
        )
        return result


def main():
    from edsl.results.Result import Result
    import json

    print("Being imported")
    json_string = """
    {
        "agent": {
            "traits": {
                "status": "Unhappy"
            }
        },
        "scenario": {
            "period": "morning"
        },
        "model": {
            "model": "gpt-3.5-turbo",
            "parameters": {
                "temperature": 0.5,
                "max_tokens": 1000,
                "top_p": 1,
                "frequency_penalty": 0,
                "presence_penalty": 0,
                "use_cache": true
            }
        },
        "iteration": 0,
        "answer": {
            "how_feeling": "Bad"
        }
    }
    """

    result = Result.from_dict(json.loads(json_string))

    result.sub_dicts
    assert result.combined_dict["how_feeling"] == "Bad"

    result.combined_dict
    assert result.get_value("answer", "how_feeling") == "Bad"

    result.key_to_data_type
    print(result)

    assert result == result.copy()

    result.to_dict()



import os
from dotenv import load_dotenv
from typing import Any
from edsl import BASE_DIR, ROOT_DIR
from edsl.exceptions import (
    InvalidEnvironmentVariableError,
    MissingEnvironmentVariableError,
)

DOTENV_PATH = os.path.join(ROOT_DIR, ".env")
CONFIG_MAP = {
    "EDSL_RUN_MODE": {
        "optional": True,
        "default": "production",
        "allowed": ["development", "production"],
        "user_message": None,
    },
    "EDSL_DATABASE_PATH": {
        "optional": True,
        "default": f"sqlite:///{os.path.join(BASE_DIR, 'data/database.db')}",
        "allowed": None,
        "user_message": None,
    },
    "EMERITUS_API_KEY": {
        "optional": True,
        "default": "local",
        "allowed": None,
        "user_message": "Please provide your Emeritus API key (https://emeritus.org/).",
    },
    "OPENAI_API_KEY": {
        "optional": False,
        "default": None,
        "allowed": None,
        "user_message": "Please provide your OpenAI API key (https://platform.openai.com/api-keys).",
    },
}


class Config:
    def __init__(self):
        self._load_env_vars_from_dotenv()
        self._set_env_vars()
        self._validate_attributes()
        self._cleanup_dotenv()

    def _load_env_vars_from_dotenv(self) -> None:
        """
        Loads environment variables from the .env file
        - Creates a .env file if it does not exist
        - .env variables override existing environment variables
        """
        if not os.path.exists(DOTENV_PATH):
            with open(DOTENV_PATH, "w") as f:
                f.write("")
        load_dotenv(dotenv_path=DOTENV_PATH, override=True)

    def _set_env_vars(self) -> None:
        """
        Sets env vars as Config class attributes.
        - optional env vars: set to default value if it exists, otherwise do not set
        - mandatory env vars: prompt the user to set it if it does not exist
        """
        for env_var, config in CONFIG_MAP.items():
            if value := os.getenv(env_var):
                setattr(self, env_var, value)
            elif default_value := config.get("default"):
                setattr(self, env_var, default_value)
                os.environ[env_var] = default_value
            elif not config.get("optional"):
                self._set_env_var(env_var, config)

    def _set_env_var(self, env_var: str, config: dict[str, Any]) -> None:
        """Attempts to set an environment variable."""
        if self.EDSL_RUN_MODE == "development":
            raise MissingEnvironmentVariableError(
                f"Missing environment variable {env_var}. Please set it:"
                f"- by adding a line `{env_var}=...` in the .env file (found at {DOTENV_PATH})"
                f"- or as a regular environment variable before running this script"
            )
        else:
            print(config.get("user_message"))
            print(f"There are three ways you can do this:")
            print(f"1. Set it as a regular environment variable")
            print(f"2. Close this program and add a line")
            print(f"{env_var}=... ")
            print(f"in the .env file (found at {DOTENV_PATH})")
            print(f"3. Enter the value below and press enter: ")
            value = input()
            value = value.strip()
            setattr(self, env_var, value)
            os.environ[env_var] = value
            with open(DOTENV_PATH, "a") as f:
                f.write(f"{env_var}={value}\n")
            print(f"Environment variable {env_var} set successfully to {value}.")
            print(f"Also saved the value in the .env file for future use.")

    def _validate_attributes(self):
        """Validates that all attributes are allowed values."""
        for attr, value in self.__dict__.items():
            config = CONFIG_MAP.get(attr)
            if config.get("allowed") and value not in config.get("allowed"):
                raise InvalidEnvironmentVariableError(
                    f"Variable {attr} has value {value}, which is not allowed.\n"
                    f"Allowed values are: {config.get('allowed')}. "
                )

    def _cleanup_dotenv(self) -> str:  # pragma: no cover
        """Prints warnings for unused and duplicate environment variables in the .env file."""
        valid_vars = set(CONFIG_MAP.keys())
        seen_vars = set()
        unused_vars = set()
        duplicate_vars = set()
        with open(DOTENV_PATH, "r") as file:
            for line in file:
                if not line.startswith("#"):
                    var = line.split("=")[0].strip()
                    if var in seen_vars:
                        duplicate_vars.add(var)
                    if var not in valid_vars:
                        unused_vars.add(var)
                    seen_vars.add(var)

        if unused_vars:
            print(f"Environment variables {unused_vars} are no longer in use.")
            print(f"- Please delete them from your .env file (found at {DOTENV_PATH}).")
            print(f"- Valid environment variables are: {valid_vars}.\n")
        if duplicate_vars:
            print(f"Environment variables {duplicate_vars} are duplicated.")
            print(
                f"- Please delete duplicates from your .env file (found at {DOTENV_PATH})."
            )

    def get(self, env_var: str) -> str:
        """Returns the value of an environment variable. If the environment variable is valid but not set, attempts to set it."""
        if env_var not in CONFIG_MAP:
            raise InvalidEnvironmentVariableError(
                f"Variable {env_var} is not a valid environment variable. "
                f"Valid environment variables are: {set(CONFIG_MAP.keys())}."
            )
        elif env_var not in self.__dict__:
            self._set_env_var(env_var, CONFIG_MAP.get(env_var))
        return self.__dict__.get(env_var)

    def show(self) -> str:
        """Prints the currently set environment vars."""
        max_env_var_length = max(len(env_var) for env_var in self.__dict__)
        print("Here are the current configuration settings:")
        for env_var, value in self.__dict__.items():
            print(f"{env_var:<{max_env_var_length}} : {value}")


# Note: Python modules are singletons. As such, once this module is imported
# the same instance of it is reused across the application.
CONFIG = Config()
